# Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
# Modifications copyright (C) 2022, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import io
import math
import os
import warnings
from datetime import datetime, timezone

import torch
from torch.utils.tensorboard import SummaryWriter

from horovod.spark.common import constants
from horovod.spark.common.util import _get_assigned_gpu_or_local_rank, to_list, _set_mp_start_method
from horovod.spark.common.store import DBFSLocalStore
from horovod.spark.torch.util import deserialize_fn

PETASTORM_HDFS_DRIVER = constants.PETASTORM_HDFS_DRIVER
METRIC_PRINT_FREQUENCY = constants.METRIC_PRINT_FREQUENCY
TOTAL_BUFFER_MEMORY_CAP_GIB = constants.TOTAL_BUFFER_MEMORY_CAP_GIB
BYTES_PER_GIB = constants.BYTES_PER_GIB
CUSTOM_SPARSE = constants.CUSTOM_SPARSE


def RemoteTrainer(estimator, metadata, last_checkpoint_state, run_id, dataset_idx):
    # Estimator parameters
    data_module = estimator.getDataModule()
    gradient_compression = estimator.getGradientCompression()
    input_shapes = estimator.getInputShapes()
    label_shapes = estimator.getLabelShapes()
    feature_columns = estimator.getFeatureCols()
    label_columns = estimator.getLabelCols()
    categorical_columns = estimator.getCategoricalCols()
    continuous_columns = estimator.getContinuousCols()
    num_labels = len(label_columns)
    should_validate = estimator.getValidation()
    batch_size = estimator.getBatchSize()
    val_batch_size = estimator.getValBatchSize() if estimator.getValBatchSize() else batch_size
    epochs = estimator.getEpochs()
    train_steps_per_epoch = estimator.getTrainStepsPerEpoch()
    validation_steps_per_epoch = estimator.getValidationStepsPerEpoch()
    sample_weight_col = estimator.getSampleWeightCol()
    metric_fn_groups = estimator.getMetrics()
    random_seed = estimator.getRandomSeed()
    user_shuffle_buffer_size = estimator.getShufflingBufferSize()
    if user_shuffle_buffer_size is not None:
        warnings.warn('shuffle_buffer_size is deprecated and will be removed in future releases, '\
                      'use shuffle instead', DeprecationWarning)
    shuffle = estimator.getShuffle()
    user_verbose = estimator.getVerbose()
    train_minibatch_fn = estimator.getTrainMinibatchFn()
    train_minibatch = train_minibatch_fn if train_minibatch_fn else _train_minibatch_fn()
    loss_fns_pre_train = to_list(estimator.getLoss(), num_labels)
    loss_constructors = to_list(estimator.getLossConstructors(), num_labels)
    transformation_fn = estimator.getTransformationFn()
    transformation = transformation_fn if transformation_fn else None
    inmemory_cache_all = estimator.getInMemoryCacheAll()
    should_use_gpu = estimator.getUseGpu()
    mp_start_method = estimator.getMpStartMethod()
    backward_passes_per_step = estimator.getBackwardPassesPerStep()

    # If loss weight is not provided, use equal loss for all the labels
    loss_weights = estimator.getLossWeights()
    if not loss_weights:
        loss_weights = [float(1) / num_labels for _ in range(num_labels)]
    else:
        if not isinstance(loss_weights, list) or \
                len(loss_weights) != len(label_columns):
            raise ValueError('loss_weights needs to be a list with the same '
                             'length as the label_columns.')

    # Data reader parameters
    train_reader_worker_count = estimator.getTrainReaderNumWorker()
    val_reader_worker_count = estimator.getValReaderNumWorker()
    reader_pool_type = estimator.getReaderPoolType()

    # Utility functions
    deserialize = deserialize_fn()
    get_optimizer_with_unscaled_lr = _get_optimizer_with_unscaled_lr_fn()
    construct_metric_value_holders = _construct_metric_value_holders_fn()
    metric_cls = _metric_cls()
    prepare_np_data = _prepare_np_data_fn()
    get_metric_avgs = _get_metric_avgs_fn()
    update_metrics = _update_metrics_fn(metric_fn_groups)
    write_metrics_summary = _write_metrics_summary_fn()
    calculate_loss = _calculate_loss_fn()

    # Storage
    store = estimator.getStore()
    remote_store = store.to_remote(run_id, dataset_idx)
    is_dbfs = isinstance(store, DBFSLocalStore)
    storage_options = store.storage_options

    def train(serialized_model, optimizer_cls, model_opt_state_serialized,
              train_rows, val_rows, avg_row_size):
        # If not empty, set it before everything else.
        if mp_start_method:
            _set_mp_start_method(mp_start_method, user_verbose)

        import torch
        import horovod.torch as hvd

        if random_seed is not None:
            torch.manual_seed(random_seed)

        # Deserializing objects
        model_opt_state = torch.load(model_opt_state_serialized)
        model = deserialize(serialized_model)

        if loss_fns_pre_train:
            loss_fns = loss_fns_pre_train
        if loss_constructors:
            local_vars = locals()
            loss_fns = [loss_constructor(**local_vars) for loss_constructor in loss_constructors]

        # Horovod: initialize library.
        hvd.init()

        if user_verbose:
            import horovod as _horovod
            print(f"Shared lib path is pointing to: {_horovod.common.process_sets._basics.MPI_LIB_CTYPES}")

        if not should_use_gpu and user_verbose:
            print("Skip pinning current process to the GPU.")

        cuda_available = torch.cuda.is_available()

        if cuda_available and not should_use_gpu:
            print("GPU is available but use_gpu is set to False."
                  "Training will proceed without GPU support.")
            cuda_available = False

        # We need to check all ranks have same device type for traning.
        # Horovod doesn't support heterogeneous allreduce for gradients.
        cuda_avail_list = hvd.allgather_object(cuda_available, name='device type')
        if cuda_avail_list.count(cuda_available) != hvd.size():
            raise RuntimeError("All ranks don't have same device type!")

        if cuda_available:
            # Horovod: pin GPU to local rank or the assigned GPU from spark.
            torch.cuda.set_device(_get_assigned_gpu_or_local_rank(local_rank=hvd.local_rank()))
            # Move model to GPU.
            model.cuda()

        # Optimizer object needs to be re-instantiated. Internally, it uses memory addresses of
        # objects as their identity and therefore it cannot be serialized and then
        # deserialized. The deserialized optimizer object stores the names of the parameters
        # with their old memory addresses but in reality those are different than the
        # reconstructed deserialized object and that creates problem.
        # Learning rate is a required parameters in SGD optimizer. It will be overridden with
        # load_state_dict.
        optimizer = optimizer_cls(model.parameters(), lr=1)
        optimizer_state = model_opt_state['optimizer']

        if last_checkpoint_state is not None:
            model.load_state_dict(last_checkpoint_state['model'])
            optimizer.load_state_dict(last_checkpoint_state['optimizer'])
        else:
            # scale the learning rate with the number of horovod workers
            for i in range(len(optimizer_state['param_groups'])):
                optimizer_state['param_groups'][i]['lr'] = \
                    optimizer_state['param_groups'][i]['lr'] * hvd.size()

            optimizer.load_state_dict(optimizer_state)

        # Horovod: broadcast parameters & optimizer state.
        hvd.broadcast_parameters(model.state_dict(), root_rank=0)
        hvd.broadcast_optimizer_state(optimizer, root_rank=0, model=model)

        dist_optimizer_args = dict(optimizer=optimizer,
                                   named_parameters=model.named_parameters())
        if gradient_compression:
            # Pass the compression arg only if it is specified by the user.
            dist_optimizer_args['compression'] = gradient_compression
        dist_optimizer_args['backward_passes_per_step'] = backward_passes_per_step
        # Horovod: wrap optimizer with DistributedOptimizer.
        optimizer = hvd.DistributedOptimizer(**dist_optimizer_args)

        # This function takes the current optimizer and constructs a new optimizer with the
        # same state except with learning rate scaled down with the number of horovod workers.
        # This is important the retraining of the model. User may retrain the model with
        # different number of workers and we need the raw learning rate to adjust with the
        # new number of workers.

        schema_fields = feature_columns + label_columns
        if sample_weight_col:
            schema_fields.append(sample_weight_col)

        if train_steps_per_epoch is None:
            steps_per_epoch = int(math.floor(float(train_rows) / batch_size / hvd.size()))
        else:
            steps_per_epoch = train_steps_per_epoch

        with remote_store.get_local_output_dir() as run_output_dir:
            logs_dir = os.path.join(run_output_dir, remote_store.logs_subdir)
            log_writer = SummaryWriter(logs_dir) if hvd.rank() == 0 else None
            ckpt_file = os.path.join(run_output_dir, remote_store.checkpoint_filename)

            def save_checkpoint():
                model.cpu()
                optimizer_with_scaled_down_lr = \
                    get_optimizer_with_unscaled_lr(hvd, optimizer, optimizer_cls, model)
                state = {
                    'model': model.state_dict(),
                    'optimizer': optimizer_with_scaled_down_lr.state_dict(),
                }
                torch.save(state, ckpt_file)
                if cuda_available:
                    model.cuda()

            if hvd.rank() == 0 and user_verbose:
                print(f"Training parameters: Epochs: {epochs}\n"
                      f"Train rows: {train_rows}, Train batch size: {batch_size}, Train_steps_per_epoch: {steps_per_epoch}\n"
                      f"Shuffle: {shuffle}, Random seed: {random_seed}\n"
                      f"Checkpoint file: {ckpt_file}, Logs dir: {logs_dir}\n")

            if validation_steps_per_epoch is None:
                validation_steps = int(math.ceil(float(val_rows) / val_batch_size / hvd.size())) if should_validate else None
            else:
                validation_steps = validation_steps_per_epoch

            data_module_kwargs = {
                'label_cols': label_columns,                    # nvtabular
                'continuous_cols': continuous_columns,          # nvtabular
                'categorical_cols': categorical_columns,        # nvtabular
                'random_seed': random_seed,                     # petastorm
                'train_dir': remote_store.train_data_path,
                'val_dir': remote_store.val_data_path,
                'num_train_epochs': epochs,
                'has_val': should_validate is not None,
                'train_batch_size': batch_size,
                'val_batch_size': val_batch_size,
                'reader_pool_type': reader_pool_type,
                'train_reader_worker_count': train_reader_worker_count,
                'val_reader_worker_count': val_reader_worker_count,
                'transform_fn': transformation_fn,
                'inmemory_cache_all': inmemory_cache_all,
                'cur_shard': hvd.rank(),
                'shard_count': hvd.size(),
                'schema_fields': schema_fields,
                'storage_options': storage_options,
                'steps_per_epoch_train': steps_per_epoch,
                'steps_per_epoch_val': validation_steps,
                'verbose': user_verbose
            }
            if user_verbose:
                print("data_module: {}".format(data_module))

            def prepare_batch(row):
                if isinstance(row, tuple) and len(row) == 2:
                    # handle iterators that return (features, label)
                    inputs = [
                        prepare_np_data(
                            row[0][col].float(), col, metadata).reshape(shape)
                        for col, shape in zip(feature_columns, input_shapes)]
                    labels = [
                        prepare_np_data(
                            row[1].float(), col, metadata)
                        for col in label_columns]
                    sample_weights = row[0].get(sample_weight_col, None)
                else:
                    inputs = [
                        prepare_np_data(
                            row[col].float(), col, metadata).reshape(shape)
                        for col, shape in zip(feature_columns, input_shapes)]
                    labels = [
                        prepare_np_data(
                            row[col].float(), col, metadata)
                        for col in label_columns]
                    sample_weights = row.get(sample_weight_col, None)

                if sample_weights is not None:
                    sample_weights = sample_weights.float()
                if cuda_available:
                    inputs = [input.cuda() for input in inputs]
                    labels = [label.cuda() for label in labels]
                    if sample_weights is not None:
                        sample_weights = sample_weights.cuda()
                return inputs, labels, sample_weights

            def transform_outputs(outputs, labels):
                if not isinstance(outputs, tuple) and not isinstance(outputs,  list):
                    outputs = [outputs]

                # reshape labels to match the output shape of the model
                if hasattr(outputs[0], 'shape'):
                    if label_shapes:
                        labels = [label.reshape(label_shape)
                                  for label, label_shape in zip(labels, label_shapes)]
                    else:
                        # If label_shapes parameter is not provided, reshape the label
                        # columns data to match the shape of the model output
                        labels = [label.reshape(output.shape) if
                                  output.shape.numel() == label.shape.numel() else label
                                  for label, output in zip(labels, outputs)]

                return outputs, labels

            def aggregate_metrics(stage, epoch, loss, metric_value_groups):
                all_metric_groups_values = get_metric_avgs(metric_value_groups)
                if remote_store.saving_runs:
                    write_metrics_summary(
                        stage, epoch, loss, all_metric_groups_values, log_writer)
                return {
                    loss.name: loss.avg.item(),
                    'all_metrics': all_metric_groups_values
                }

            def loss_fn(outputs, labels, sample_weights):
                loss = calculate_loss(outputs, labels, loss_weights, loss_fns, sample_weights)
                return loss

            def print_metrics(batch_idx, loss, metric_value_groups, phase):
                if user_verbose > 0 and hvd.rank() == 0 and \
                        batch_idx % METRIC_PRINT_FREQUENCY == 0:
                    print("{phase}\tepoch:\t{epoch}\tstep\t{batch_idx}:\t{metrics}".
                          format(phase=phase,
                                 epoch=epoch,
                                 batch_idx=batch_idx,
                                 metrics=aggregate_metrics(phase, epoch, loss,
                                                           metric_value_groups)))

            def _train(epoch, train_loader_iter):
                model.train()
                train_loss = metric_cls('loss', hvd)
                metric_value_groups = construct_metric_value_holders(
                    metric_cls, metric_fn_groups, label_columns, hvd)

                # iterate on one epoch
                for batch_idx in range(steps_per_epoch):
                    row = next(train_loader_iter)
                    inputs, labels, sample_weights = prepare_batch(row)
                    outputs, loss = train_minibatch(model, optimizer, transform_outputs,
                                                    loss_fn, inputs, labels, sample_weights,
                                                    backward_passes_per_step, batch_idx)
                    update_metrics(metric_value_groups, outputs, labels)
                    train_loss.update(loss)
                    print_metrics(batch_idx, train_loss, metric_value_groups, 'train')
                optimizer.step()

                return aggregate_metrics('train', epoch, train_loss, metric_value_groups)

            if should_validate:
                if hvd.rank() == 0 and user_verbose:
                    print(f"Val rows: {val_rows}, Val batch size: {val_batch_size}, Val_steps_per_epoch: {validation_steps}\n")

                def _validate(epoch, val_loader_iter):
                    model.eval()
                    val_loss = metric_cls('loss', hvd)

                    metric_value_groups = construct_metric_value_holders(
                        metric_cls, metric_fn_groups, label_columns, hvd)

                    # iterate on one epoch
                    for batch_idx in range(validation_steps):
                        row = next(val_loader_iter)
                        inputs, labels, sample_weights = prepare_batch(row)

                        outputs = model(*inputs)
                        outputs, labels = transform_outputs(outputs, labels)

                        loss = calculate_loss(
                            outputs, labels, loss_weights, loss_fns, sample_weights)
                        val_loss.update(loss)
                        update_metrics(metric_value_groups, outputs, labels)
                        print_metrics(batch_idx, val_loss, metric_value_groups, 'val')
                    return aggregate_metrics('val', epoch, val_loss, metric_value_groups)

            history = []

            with data_module(**data_module_kwargs) as dm:
                train_loader_iter = iter(dm.train_data())
                if should_validate:
                    val_loader_iter = iter(dm.val_data())

                for epoch in range(epochs):

                    epoch_metrics = {
                        'epoch': epoch,
                        'train': _train(epoch, train_loader_iter)
                    }

                    if should_validate:
                        epoch_metrics['validation'] = _validate(epoch, val_loader_iter)

                    if user_verbose > 0:
                        pdt_dt = datetime.now(timezone.utc)
                        pdt_time_str = pdt_dt.strftime("%Y-%b-%d %H:%M:%S UTC")
                        print(pdt_time_str, epoch_metrics)

                    history.append(epoch_metrics)
                    if hvd.rank() == 0:
                        # Save model after every epoch
                        save_checkpoint()
                        if remote_store.saving_runs:
                            remote_store.sync(run_output_dir)

            if hvd.rank() == 0:
                best_checkpoint = torch.load(ckpt_file)
                serialized_checkpoint = io.BytesIO()
                torch.save(best_checkpoint, serialized_checkpoint)
                serialized_checkpoint.seek(0)
                return history, serialized_checkpoint

    return train


def _train_minibatch_fn():
    def train_minibatch(model, optimizer, transform_outputs, loss_fn, inputs, labels, sample_weights, backward_passes_per_step, batch_idx):
        if batch_idx % backward_passes_per_step == 0:
            if batch_idx != 0:
                optimizer.step()
            optimizer.zero_grad()
        outputs = model(*inputs)
        outputs, labels = transform_outputs(outputs, labels)
        loss = loss_fn(outputs, labels, sample_weights)
        if backward_passes_per_step > 1:
            loss.div_(float(backward_passes_per_step))
        loss.backward()
        return outputs, loss
    return train_minibatch


def _get_optimizer_with_unscaled_lr_fn():
    def get_optimizer_with_unscaled_lr(hvd, current_optimizer, optimizer_cls, model):
        optimizer_state = current_optimizer.state_dict()
        # scale down the learning rate with the number of horovod workers
        for i in range(len(optimizer_state['param_groups'])):
            optimizer_state['param_groups'][i]['lr'] = \
                optimizer_state['param_groups'][i]['lr'] / hvd.size()
        optimizer = optimizer_cls(model.parameters(), lr=1)
        optimizer.load_state_dict(optimizer_state)
        return optimizer

    return get_optimizer_with_unscaled_lr


def _construct_metric_value_holders_fn():
    def construct_metric_value_holders(metric_class, metric_fn_groups, label_columns, hvd):
        metric_values = []
        for group_number, metric_group in enumerate(metric_fn_groups):
            metric_group_val = []
            for label_col in label_columns:
                metric_group_val.append(
                    metric_class('group_' + str(group_number) + '_' + label_col, hvd))

            metric_values.append(metric_group_val)
        return metric_values
    return construct_metric_value_holders


def _metric_cls():
    # Horovod: average metrics from distributed training.
    class Metric(object):
        def __init__(self, name, hvd):
            self.name = name
            self.sum = torch.tensor(0.)
            self.n = torch.tensor(0.)
            self.hvd = hvd

        def update(self, val):
            self.sum += self.hvd.allreduce(val.detach().cpu(), name=self.name)
            self.n += 1

        @property
        def avg(self):
            return self.sum / self.n

    return Metric


def _prepare_np_data_fn():
    def prepare_np_data(rows, col_name, metadata):
        intermediate_format = metadata[col_name]['intermediate_format']
        if intermediate_format != CUSTOM_SPARSE:
            return rows

        shape = metadata[col_name]['shape']
        num_rows = rows.shape[0]
        dense_rows = torch.zeros([num_rows, shape])
        for r in range(num_rows):
            size = rows[r][0].long()
            dense_rows[r][rows[r][1:size + 1].long()] = \
                rows[r][size + 1:2 * size + 1]
        return dense_rows

    return prepare_np_data


def _get_metric_avgs_fn():
    def get_metric_avgs(metric_value_groups):
        all_metric_groups_values = []
        for metric_value_group in metric_value_groups:
            metric_avgs = {}
            for metric in metric_value_group:
                metric_avgs[metric.name] = metric.avg.item()
            all_metric_groups_values.append(metric_avgs)
        return all_metric_groups_values

    return get_metric_avgs


def _update_metrics_fn(metric_fn_groups):
    def update_metrics(metric_value_groups, outputs, labels):
        """
        metric_value_groups is a list of metric functions. For example, for a model with 3
        outputs, we can define these two metric groups
        [
            [metric_fn1],
            [metric_fn21,metric_fn22,metric_fn23],
        ]

        In this example, first metric group provides only one metric function. This
        function will be used to calculate the metric on all of the model outputs. Second
        metric groups, however, defines one metric function per output.
        """

        num_outputs = len(outputs)
        for metric_fn_group, metric_value_group in zip(metric_fn_groups, metric_value_groups):
            if len(metric_fn_group) == 1:
                _metric_fn_group = [metric_fn_group[0] for _ in range(num_outputs)]
            else:
                _metric_fn_group = metric_fn_group

            for metric_val, metric_fn, output_group, label_group in \
                    zip(metric_value_group, _metric_fn_group, outputs, labels):
                metric_val.update(metric_fn(output_group, label_group))

        return metric_value_groups

    return update_metrics


def _write_metrics_summary_fn():
    def write_metrics_summary(stage, epoch, loss_metric, metric_value_groups, log_writer):
        if not log_writer:
            return

        log_writer.add_scalar('{}/{}'.format(stage, loss_metric.name),
                              loss_metric.avg.item(), epoch)

        for idx, metric_value_group in enumerate(metric_value_groups):
            for metric in metric_value_group:
                log_writer.add_scalar('{}/{}:{}'.format(stage, metric.name, idx),
                                      metric.avg.item(), epoch)

    return write_metrics_summary


def _calculate_loss_fn():
    def calculate_loss(outputs, labels, loss_weights, loss_fns, sample_weights=None):
        if sample_weights is not None:
            # when reduction='none', loss function returns the value of all the losses
            # from all the samples. We multiply each sample's weight to its loss and
            # then take the mean of the weight adjusted losses from all the samples in the
            # batch. Note that this approach is not "weighted average" because the sum of
            # the sample weights in each batch does not necessarily add up to one. If we add
            # the weights and divide the sum to the sum of weights, the impact of two
            # samples with identical weights but in different batches will not be equal on
            # the calculated gradients.
            losses = []
            for output, label, loss_fn, loss_weight in zip(outputs, labels,
                                                           loss_fns, loss_weights):
                weight_adjusted_sample_losses = \
                    loss_fn(output, label, reduction='none').flatten() * sample_weights
                output_loss = weight_adjusted_sample_losses.mean()
                losses.append(output_loss * loss_weight)
        else:
            losses = [loss_fn(output, label) * loss_weight for
                      output, label, loss_fn, loss_weight in
                      zip(outputs, labels, loss_fns, loss_weights)]

        loss = sum(losses)
        return loss

    return calculate_loss



# Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


class BaseSettings(object):
    def __init__(self, num_proc=None, verbose=0, ssh_port=None, ssh_identity_file=None, extra_mpi_args=None,
                 tcp_flag=None, binding_args=None, key=None, start_timeout=None, output_filename=None,
                 run_func_mode=None, nics=None, elastic=False, prefix_output_with_timestamp=False):
        """
        :param num_proc: number of horovod processes (-np)
        :type num_proc: int
        :param verbose: level of verbosity
        :type verbose: int
        :param ssh_port: SSH port on all the hosts
        :type ssh_port: int
        :param ssh_identity_file: SSH identity (private key) file
        :type ssh_identity_file: string
        :param extra_mpi_args: Extra MPI arguments to pass to mpirun
        :type extra_mpi_args: string
        :param tcp_flag: TCP only communication flag
        :type tcp_flag: boolean
        :param binding_args: Process binding arguments
        :type binding_args: string
        :param key: used for encryption of parameters passed across the hosts
        :type key: str
        :param start_timeout: has to finish all the checks before this timeout runs out.
        :type start_timeout: horovod.runner.common.util.timeout.Timeout
        :param output_filename: optional filename to redirect stdout / stderr by process
        :type output_filename: string
        :param run_func_mode: whether it is run function mode
        :type run_func_mode: boolean
        :param nics: specify the NICs to be used for tcp network communication.
        :type nics: Iterable[str]
        :param elastic: enable elastic auto-scaling and fault tolerance mode
        :type elastic: boolean
        :param prefix_output_with_timestamp: shows timestamp in stdout/stderr forwarding on the driver
        :type prefix_output_with_timestamp: boolean
        """
        self.num_proc = num_proc
        self.verbose = verbose
        self.ssh_port = ssh_port
        self.ssh_identity_file = ssh_identity_file
        self.extra_mpi_args = extra_mpi_args
        self.tcp_flag = tcp_flag
        self.binding_args = binding_args
        self.key = key
        self.start_timeout = start_timeout
        self.output_filename = output_filename
        self.run_func_mode = run_func_mode
        self.nics = nics
        self.elastic = elastic
        self.prefix_output_with_timestamp = prefix_output_with_timestamp


class Settings(BaseSettings):
    def __init__(self, hosts=None, **kwargs):
        """
        :param hosts: string, comma-delimited, of hostname[s] with slots number[s]
        :type hosts: string
        """
        super(Settings, self).__init__(**kwargs)
        self.hosts = hosts



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Modifications copyright (c) 2020, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from collections import defaultdict

# Load all the necessary MXNet C types.
import ctypes
import os

import mxnet as mx
from mxnet.base import c_handle_array, c_str, c_str_array, check_call, string_types

from horovod.common.util import check_installed_version, get_ext_suffix, get_average_backwards_compatibility_fun
from horovod.common.basics import HorovodBasics as _HorovodBasics
from horovod.common.process_sets import _setup as _setup_process_sets
from horovod.common.process_sets import ProcessSet, global_process_set, add_process_set, remove_process_set, \
    _temp_process_set_object

# Check possible symbol not found error from mxnet version mismatch
try:
    _basics = _HorovodBasics(__file__, 'mpi_lib')
except Exception as e:
    check_installed_version('mxnet', mx.__version__, e)
    raise e
else:
    check_installed_version('mxnet', mx.__version__)

# import basic methods
shutdown = _basics.shutdown
is_initialized = _basics.is_initialized
start_timeline = _basics.start_timeline
stop_timeline = _basics.stop_timeline
size = _basics.size
local_size = _basics.local_size
cross_size = _basics.cross_size
rank = _basics.rank
local_rank = _basics.local_rank
cross_rank = _basics.cross_rank
mpi_threads_supported = _basics.mpi_threads_supported
mpi_enabled = _basics.mpi_enabled
mpi_built = _basics.mpi_built
gloo_enabled = _basics.gloo_enabled
gloo_built = _basics.gloo_built
nccl_built = _basics.nccl_built
ddl_built = _basics.ddl_built
ccl_built = _basics.ccl_built
cuda_built = _basics.cuda_built
rocm_built = _basics.rocm_built

def init(*args, **kwargs):
    _basics.init(*args, **kwargs)
    # Call set up again to make sure the basics is in sync
    _setup_process_sets(_basics)

# import reduction op values
Average = _basics.Average
Sum = _basics.Sum
Adasum = _basics.Adasum
Min = _basics.Min
Max = _basics.Max
Product = _basics.Product

handle_average_backwards_compatibility = get_average_backwards_compatibility_fun(_basics)

dll_path = os.path.join(os.path.dirname(__file__),
                        'mpi_lib' + get_ext_suffix())
MPI_MXNET_LIB_CTYPES = ctypes.CDLL(dll_path, ctypes.RTLD_GLOBAL)

_setup_process_sets(_basics)


def allreduce(tensor, average=None, name=None, priority=0, prescale_factor=1.0,
              postscale_factor=1.0, process_set=global_process_set, op=None):
    """
    A function that performs averaging or summation of the input tensor over
    all the Horovod processes. The input tensor is not modified.

    The reduction operation is keyed by the name. If name is not provided, an
    incremented auto-generated name is used. The tensor type and shape must be
    the same on all Horovod processes for a given name. The reduction will not
    start until all processes are ready to send and receive the tensor.

    This acts as a thin wrapper around an autograd function.  If your input
    tensor requires gradients, then callings this function will allow gradients
    to be computed and backpropagated.

    Arguments:
        tensor: A tensor to average or sum.
        average:
            .. warning:: .. deprecated:: 0.24.0

                Use `op` instead. Will be removed in v1.0.

        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        name: A name of the reduction operation.
        priority: The priority of this operation. Higher priority operations
                  are likely to be executed before other operations.
        prescale_factor: Multiplicative factor to scale tensor before allreduce
        postscale_factor: Multiplicative factor to scale tensor after allreduce
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A tensor of the same shape and type as `tensor`, averaged or summed
        across all processes.
    """
    op = handle_average_backwards_compatibility(op, average)
    assert op != Adasum

    output = mx.nd.zeros(shape=tensor.shape, ctx=tensor.context,
                         dtype=tensor.dtype)

    c_in = tensor.handle
    c_out = output.handle
    c_name = c_str(name) if isinstance(name, string_types) else ctypes.c_char_p(None)

    check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_allreduce_async(
        ctypes.byref(c_in), ctypes.byref(c_out), c_name, ctypes.c_int(op),
        ctypes.c_int(priority),
        ctypes.c_double(prescale_factor),
        ctypes.c_double(postscale_factor),
        ctypes.c_int(1),
        ctypes.c_int(process_set.process_set_id)))

    return output


def allreduce_(tensor, average=None, name=None, priority=0, prescale_factor=1.0,
               postscale_factor=1.0, process_set=global_process_set,
               op=None):
    """
    A function that performs in-place averaging or summation of the input
    tensor over all the Horovod processes.

    The reduction operation is keyed by the name. If name is not provided, an
    incremented auto-generated name is used. The tensor type and shape must be
    the same on all Horovod processes for a given name. The reduction will not
    start until all processes are ready to send and receive the tensor.

    Arguments:
        tensor: A tensor to average or sum.
        average:
            .. warning:: .. deprecated:: 0.24.0

                Use `op` instead. Will be removed in v1.0.

        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        name: A name of the reduction operation.
        priority: The priority of this operation. Higher priority operations
                  are likely to be executed before other operations.
        prescale_factor: Multiplicative factor to scale tensor before allreduce
        postscale_factor: Multiplicative factor to scale tensor after allreduce
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A tensor of the same shape and type as `tensor`, averaged or summed
        across all processes.
    """
    op = handle_average_backwards_compatibility(op, average)
    assert op != Adasum

    c_in = tensor.handle
    c_out = tensor.handle
    c_name = c_str(name) if isinstance(name, string_types) else ctypes.c_char_p(None)

    check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_allreduce_async(
        ctypes.byref(c_in), ctypes.byref(c_out), c_name, ctypes.c_int(op),
        ctypes.c_int(priority),
        ctypes.c_double(prescale_factor),
        ctypes.c_double(postscale_factor),
        ctypes.c_int(1),
        ctypes.c_int(process_set.process_set_id)))

    return tensor


def grouped_allreduce(tensors, average=None, name=None, priority=0, prescale_factor=1.0,
                      postscale_factor=1.0, process_set=global_process_set, op=None):
    """
    A function that performs averaging or summation of the input
    tensors over all the Horovod processes. The input tensors are not modified.

    The reduction operations are keyed by the base name. If a base name is not
    provided, an incremented auto-generated base name is used. Reductions are
    performed across tensors in the same list position. The tensor type and
    shape must be the same on all Horovod processes for tensors sharing
    positions in the input tensor list. The reduction will not start until all
    processes are ready to send and receive the tensors.

    Arguments:
        tensors: A list of tensors to average or sum.
        average:
            .. warning:: .. deprecated:: 0.24.0

                Use `op` instead. Will be removed in v1.0.

        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        name: A base name to use for the group reduction operation
        priority: The priority of this operation. Higher priority operations
                  are likely to be executed before other operations.
        prescale_factor: Multiplicative factor to scale tensor before allreduce
        postscale_factor: Multiplicative factor to scale tensor after allreduce
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A list containing tensors of the same shape and type as in `tensors`,
        averaged or summed across all processes.
    """
    op = handle_average_backwards_compatibility(op, average)
    assert op != Adasum

    if not tensors:
      return tensors

    outputs = [mx.nd.zeros(shape=tensor.shape, ctx=tensor.context,
                         dtype=tensor.dtype) for tensor in tensors]

    c_in = c_handle_array(tensors)
    c_out = c_handle_array(outputs)
    c_name = c_str(name) if isinstance(name, string_types) else ctypes.c_char_p(None)

    check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_allreduce_async(
        c_in, c_out, c_name, ctypes.c_int(op),
        ctypes.c_int(priority),
        ctypes.c_double(prescale_factor),
        ctypes.c_double(postscale_factor),
        ctypes.c_int(len(tensors)),
        ctypes.c_int(process_set.process_set_id)))

    return outputs


def grouped_allreduce_(tensors, average=None, name=None, priority=0, prescale_factor=1.0,
                       postscale_factor=1.0, process_set=global_process_set, op=None):
    """
    A function that performs in-place averaging or summation of the input
    tensors over all the Horovod processes.

    The reduction operations are keyed by the base name. If a base name is not
    provided, an incremented auto-generated base name is used. Reductions are
    performed across tensors in the same list position. The tensor type and
    shape must be the same on all Horovod processes for tensors sharing
    positions in the input tensor list. The reduction will not start until all
    processes are ready to send and receive the tensors.

    Arguments:
        tensors: A list of tensors to average or sum.
        average:
            .. warning:: .. deprecated:: 0.24.0

                Use `op` instead. Will be removed in v1.0.

        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        name: A base name to use for the group reduction operation
        priority: The priority of this operation. Higher priority operations
                  are likely to be executed before other operations.
        prescale_factor: Multiplicative factor to scale tensor before allreduce
        postscale_factor: Multiplicative factor to scale tensor after allreduce
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A list containing tensors of the same shape and type as in `tensors`,
        averaged or summed across all processes.
    """
    op = handle_average_backwards_compatibility(op, average)
    assert op != Adasum

    if not tensors:
      return tensors

    c_in = c_handle_array(tensors)
    c_out = c_handle_array(tensors)
    c_name = c_str(name) if isinstance(name, string_types) else ctypes.c_char_p(None)

    check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_allreduce_async(
        c_in, c_out, c_name, ctypes.c_int(op),
        ctypes.c_int(priority),
        ctypes.c_double(prescale_factor),
        ctypes.c_double(postscale_factor),
        ctypes.c_int(len(tensors)),
        ctypes.c_int(process_set.process_set_id)))

    return tensors


def allgather(tensor, name=None, priority=0, process_set=global_process_set):
    """
    A function that concatenates the input tensor with the same input tensor on
    all other Horovod processes. The input tensor is not modified.

    The concatenation is done on the first dimension, so the input tensors on
    the different processes must have the same rank and shape, except for the
    first dimension, which is allowed to be different.

    Arguments:
        tensor: A tensor to allgather.
        name: A name of the allgather operation.
        priority: The priority of this operation. Higher priority operations
                  are likely to be executed before other operations.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A tensor of the same type as `tensor`, concatenated on dimension zero
        across all processes. The shape is identical to the input shape, except
        for the first dimension, which may be greater and is the sum of all
        first dimensions of the tensors in different Horovod processes.
    """
    assert(isinstance(tensor, mx.nd.NDArray))
    # Size of output is unknown, create output array that
    # will be resized during Horovod operation
    output = mx.nd.empty(shape=(1,), ctx=tensor.context,
                         dtype=tensor.dtype)
    c_in = tensor.handle
    c_out = output.handle
    c_name = c_str(name) if isinstance(name, string_types) else ctypes.c_char_p(None)

    check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_allgather_async(
        ctypes.byref(c_in), ctypes.byref(c_out), c_name, ctypes.c_int(priority),
        ctypes.c_int(process_set.process_set_id), ctypes.c_int(1)
        ))

    # Need to block here so changes to output tensor are visible
    output.wait_to_read()
    return output


def grouped_allgather(tensors, name=None, priority=0, process_set=global_process_set):
    """
    A function that concatenates each input tensor with the corresponding
    input tensor on all other Horovod processes for a list of input tensors.
    The input tensors are not modified.

    The concatenation is done on the first dimension, so the corresponding input
    tensors on the different processes must have the same rank and shape, except
    for the first dimension, which is allowed to be different.

    Arguments:
        tensors: A list of tensors to allgather.
        name: A base name to use for the group allgather operation.
        priority: The priority of this operation. Higher priority operations
                  are likely to be executed before other operations.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A list containing tensors of the same type as in `tensors`. Each tensor
        is concatenated on dimension zero across all processes. Its shape is
        identical to the corresponding input shape, expect for the first
        dimension, which may be greater and is the sum of all first dimensions
        of the corresponding tensor in different Horovod processes.
    """
    assert(all(isinstance(t, mx.nd.NDArray) for t in tensors))
    # Sizes of outputs are unknown, create output arrays that
    # will be resized during Horovod operation
    outputs = [mx.nd.empty(shape=(1,), ctx=t.context, dtype=t.dtype)
               for t in tensors]
    c_in = c_handle_array(tensors)
    c_out = c_handle_array(outputs)
    c_name = c_str(name) if isinstance(name, string_types) else ctypes.c_char_p(None)

    check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_allgather_async(
        c_in, c_out, c_name, ctypes.c_int(priority),
        ctypes.c_int(process_set.process_set_id),
        ctypes.c_int(len(tensors))))

    # Need to block here so changes to output tensors are visible
    for o in outputs:
        o.wait_to_read()
    return outputs


def broadcast(tensor, root_rank, name=None, priority=0, process_set=global_process_set):
    """
    A function that broadcasts the input tensor on root rank to the same input
    tensor on all other Horovod processes. The input tensor is not modified.

    The broadcast operation is keyed by the name. If name is not provided, an
    incremented auto-generated name is used. The tensor type and shape must be
    the same on all Horovod processes for a given name. The broadcast will not
    start until all processes are ready to send and receive the tensor.

    This acts as a thin wrapper around an autograd function.  If your input
    tensor requires gradients, then callings this function will allow gradients
    to be computed and backpropagated.

    Arguments:
        tensor: A tensor to broadcast.
        root_rank: The rank to broadcast the value from.
        name: A name of the broadcast operation.
        priority: The priority of this operation. Higher priority operations
                  are likely to be executed before other operations.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A tensor of the same shape and type as `tensor`, with the value
        broadcasted from root rank.
    """
    if rank() == root_rank:
        output = tensor.copy()
    else:
        output = mx.nd.zeros(shape=tensor.shape, ctx=tensor.context,
                             dtype=tensor.dtype)
    c_in = tensor.handle
    c_out = output.handle
    if isinstance(name, string_types):
        check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_broadcast_async(
            c_in, c_out, c_str(name), ctypes.c_int(root_rank),
            ctypes.c_int(priority), ctypes.c_int(process_set.process_set_id)))
    else:
        check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_broadcast_async(
            c_in, c_out, name, ctypes.c_int(root_rank),
            ctypes.c_int(priority), ctypes.c_int(process_set.process_set_id)))
    return output


def broadcast_(tensor, root_rank, name=None, priority=0, process_set=global_process_set):
    """
    A function that broadcasts the input tensor on root rank to the same input
    tensor on all other Horovod processes. The operation is performed in-place.

    The broadcast operation is keyed by the name. If name is not provided, an
    incremented auto-generated name is used. The tensor type and shape must be
    the same on all Horovod processes for a given name. The broadcast will not
    start until all processes are ready to send and receive the tensor.

    Arguments:
        tensor: A tensor to broadcast.
        root_rank: The rank to broadcast the value from.
        name: A name of the broadcast operation.
        priority: The priority of this operation. Higher priority operations
                  are likely to be executed before other operations.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.
    Returns:
        A tensor of the same shape and type as `tensor`, with the value
        broadcasted from root rank.
    """
    c_in = tensor.handle
    c_out = tensor.handle
    if isinstance(name, string_types):
        check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_broadcast_async(
            c_in, c_out, c_str(name), ctypes.c_int(root_rank),
            ctypes.c_int(priority), ctypes.c_int(process_set.process_set_id)))
    else:
        check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_broadcast_async(
            c_in, c_out, name, ctypes.c_int(root_rank),
            ctypes.c_int(priority), ctypes.c_int(process_set.process_set_id)))
    return tensor

def alltoall(tensor, splits=None, name=None, priority=0, process_set=global_process_set):
    """
    A function that scatters slices of the input tensor to all other Horovod processes
    and returns a tensor of gathered slices from all other Horovod processes. The input
    tensor is not modified.

    The slicing is done on the first dimension, so the input tensors on
    the different processes must have the same rank and shape, except for the
    first dimension, which is allowed to be different.

    Arguments:
        tensor: A tensor to distribute with alltoall.
        splits: A tensor of integers in rank order describing how many
                elements in `tensor` to send to each worker.  Splitting is
                applied along the first dimension of `tensor`. If `splits` is
                not provided, the first dimension is split equally by the
                number of Horovod processes.
        name: A name of the alltoall operation.
        priority: The priority of this operation. Higher priority operations
                  are likely to be executed before other operations.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        1) A tensor containing the gathered tensor data from all workers.
        2) If `splits` has been provided: A tensor of integers in rank order
           describing how many elements in the output tensor have been received
           from each worker.
    """
    assert(isinstance(tensor, mx.nd.NDArray))

    should_return_received_splits = (splits is not None)
    if splits is None:
        # If splits not provided, create empty tensor as placeholder
        splits = mx.nd.array([], ctx=mx.cpu(), dtype='int32')
    elif not isinstance(splits, mx.nd.NDArray):
        splits = mx.nd.array(splits, ctx=mx.cpu(), dtype='int32')

    # Size of output is unknown, create output array that
    # will be resized during Horovod operation
    output = mx.nd.empty(shape=(1,), ctx=tensor.context,
                         dtype=tensor.dtype)
    output_received_splits = mx.nd.empty(shape=(process_set.size(),), ctx=mx.cpu(), dtype='int32')
    c_in = tensor.handle
    c_out = output.handle
    c_splits = splits.handle
    c_out_recv_splits = output_received_splits.handle
    if isinstance(name, string_types):
        check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_alltoall_async(
            c_in, c_out, c_str(name), c_splits, c_out_recv_splits, ctypes.c_int(priority),
            ctypes.c_int(process_set.process_set_id)))
    else:
        check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_alltoall_async(
            c_in, c_out, name, c_splits, c_out_recv_splits, ctypes.c_int(priority),
            ctypes.c_int(process_set.process_set_id)))

    # Need to block here so changes to output tensor are visible
    output.wait_to_read()
    if should_return_received_splits:
        output_received_splits.wait_to_read()
        return output, output_received_splits
    else:
        return output


def reducescatter(tensor, op=Average, name=None, priority=0,
                  process_set=global_process_set, prescale_factor=1.0, postscale_factor=1.0):
    """
    A function that performs asynchronous averaging or summation of the input tensor
    over all the Horovod processes, then scatters the results across all Horovod
    processes. The input tensor is not modified.

    The reduction operation is keyed by the name. If name is not provided, an
    incremented auto-generated name is used. The tensor type and shape must be
    the same on all Horovod processes for a given name. The reduction will not
    start until all processes are ready to send and receive the tensor.

    This acts as a thin wrapper around an autograd function.  If your input
    tensor requires gradients, then callings this function will allow gradients
    to be computed and backpropagated.

    Arguments:
        tensor: A tensor to average/sum and scatter.
        op: The reduction operation to combine tensors across different ranks.
            Can be Average (default) or Sum.
        name: A name of the reduction operation.
        priority: The priority of this operation. Higher priority operations
                  are likely to be executed before other operations.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.
        prescale_factor: Multiplicative factor to scale tensor before reducescatter.
        postscale_factor: Multiplicative factor to scale tensor after reducescatter.

    Returns:
        A tensor of the same rank and type as `tensor` across all processes.
        The shape is identical to the input shape except for the first dimension,
        which will be divided across the different Horovod processes.
    """
    assert(isinstance(tensor, mx.nd.NDArray))
    assert(op in [Average, Sum])
    if tensor.shape == ():
        raise ValueError("reducescatter does not support scalar inputs")
    # Size of output is unknown, create output array that
    # will be resized during Horovod operation
    output = mx.nd.empty(shape=(1,), ctx=tensor.context,
                         dtype=tensor.dtype)
    c_in = tensor.handle
    c_out = output.handle
    c_name = c_str(name) if isinstance(name, string_types) else ctypes.c_char_p(None)

    check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_reducescatter_async(
        ctypes.byref(c_in), ctypes.byref(c_out), c_name, ctypes.c_int(priority),
        ctypes.c_int(process_set.process_set_id), ctypes.c_int(1),
        ctypes.c_int(op), ctypes.c_double(prescale_factor), ctypes.c_double(postscale_factor))
    )

    # Need to block here so changes to output tensor are visible
    output.wait_to_read()

    return output


def grouped_reducescatter(tensors, op=Average, name=None, priority=0,
                          process_set=global_process_set, prescale_factor=1.0, postscale_factor=1.0):
    """
    A function that performs reduction of a list of input tensors over all the
    Horovod processes, then scatters the results across all Horovod processes. The
    input tensors are not modified.

    The reduction operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The reduction will not start until all processes
    are ready to send and receive the tensor.

    This acts as a thin wrapper around an autograd function.  If your input
    tensor requires gradients, then callings this function will allow gradients
    to be computed and backpropagated.

    Arguments:
        tensors: A list of tensors to average and sum.
        op: The reduction operation to combine tensors across different ranks.
            Can be Average (default) or Sum.
        name: A base name to use for the group reduction operation.
        priority: The priority of this operation. Higher priority operations
                  are likely to be executed before other operations.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.
        prescale_factor: Multiplicative factor to scale tensors before reducescatter.
        postscale_factor: Multiplicative factor to scale tensors after reducescatter.

    Returns:
        A list containing tensors of the same rank and type as in `tensors`. For each
        tensor the shape is identical to the input shape, except for the first dimension,
        which will be divided across the different Horovod processes.
    """
    assert(all(isinstance(t, mx.nd.NDArray) for t in tensors))
    assert(op in [Average, Sum])
    if any(tensor.shape == () for tensor in tensors):
        raise ValueError("groued_reducescatter does not support scalar inputs")
    # Sizes of outputs are unknown, create output arrays that
    # will be resized during Horovod operation
    outputs = [mx.nd.empty(shape=(1,), ctx=t.context, dtype=t.dtype)
               for t in tensors]
    c_in = c_handle_array(tensors)
    c_out = c_handle_array(outputs)
    c_name = c_str(name) if isinstance(name, string_types) else ctypes.c_char_p(None)

    check_call(MPI_MXNET_LIB_CTYPES.horovod_mxnet_reducescatter_async(
        c_in, c_out, c_name, ctypes.c_int(priority),
        ctypes.c_int(process_set.process_set_id), ctypes.c_int(len(tensors)),
        ctypes.c_int(op), ctypes.c_double(prescale_factor), ctypes.c_double(postscale_factor)))

    # Need to block here so changes to output tensors are visible
    for o in outputs:
        o.wait_to_read()

    return outputs



# Copyright IBM Corp. 2020. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import io
import os

import yaml

from horovod.common.util import _cache
from horovod.runner.common.util import safe_shell_exec
from horovod.runner.util.remote import get_remote_command


class LSFUtils:
    """LSF Utilities"""
    _CSM_ALLOCATION_QUERY = "/opt/ibm/csm/bin/csm_allocation_query"
    _CSM_NODE_QUERY = "/opt/ibm/csm/bin/csm_node_attributes_query"
    _LSCPU_CMD = "LANG=en_US.utf8 lscpu"
    _THREAD_KEY= "Thread(s) per core"
    _csm_allocation_info = {}

    @staticmethod
    def using_lsf():
        """Returns True if LSF was used to start the current process."""
        return "LSB_JOBID" in os.environ

    @staticmethod
    def get_allocation_info():
        """Returns and sets the static CSM allocation info."""
        if not LSFUtils._csm_allocation_info:
            lsf_allocation_id = os.environ["CSM_ALLOCATION_ID"].strip()
            output = io.StringIO()
            exit_code = safe_shell_exec.execute("{cmd} -a {allocation}".format(
                cmd=LSFUtils._CSM_ALLOCATION_QUERY, allocation=lsf_allocation_id),
                stdout=output, stderr=output)
            if exit_code != 0:
                raise RuntimeError(
                    "{cmd} failed with exit code {exit_code}".format(
                        cmd=LSFUtils._CSM_ALLOCATION_QUERY, exit_code=exit_code))
            LSFUtils._csm_allocation_info = yaml.safe_load(output.getvalue())
            # Fetch the total number of cores and gpus for the first host
            output = io.StringIO()
            exit_code = safe_shell_exec.execute("{cmd} -n {node}".format(
                cmd=LSFUtils._CSM_NODE_QUERY,
                node=LSFUtils._csm_allocation_info["compute_nodes"][0]),
                stdout=output, stderr=output)
            if exit_code != 0:
                raise RuntimeError(
                    "{cmd} failed with exit code {exit_code}".format(
                        cmd=LSFUtils._CSM_NODE_QUERY, exit_code=exit_code))
            node_output = yaml.safe_load(output.getvalue())
            total_core_count = (int(node_output["Record_1"]["discovered_cores"]) -
                               int(node_output["Record_1"]["discovered_sockets"]) * LSFUtils._csm_allocation_info["isolated_cores"])
            LSFUtils._csm_allocation_info["compute_node_cores"]= total_core_count
            LSFUtils._csm_allocation_info["compute_node_gpus"] = int(node_output["Record_1"]["discovered_gpus"])
            # Sorting LSF hostnames
            LSFUtils._csm_allocation_info["compute_nodes"].sort()
        return LSFUtils._csm_allocation_info

    @staticmethod
    def get_compute_hosts():
        """Returns the list of LSF compute hosts."""
        return LSFUtils.get_allocation_info()["compute_nodes"]

    @staticmethod
    def get_num_cores():
        """Returns the number of cores per node."""
        return LSFUtils.get_allocation_info()["compute_node_cores"]

    @staticmethod
    def get_num_gpus():
        """Returns the number of gpus per node."""
        return LSFUtils.get_allocation_info()["compute_node_gpus"]

    @staticmethod
    @_cache
    def get_num_processes():
        """Returns the total number of processes."""
        return len(LSFUtils.get_compute_hosts()) * LSFUtils.get_num_gpus()

    @staticmethod
    @_cache
    def get_num_threads():
        """Returns the number of hardware threads."""
        lscpu_cmd = get_remote_command(LSFUtils._LSCPU_CMD, host=LSFUtils.get_compute_hosts()[0])
        output = io.StringIO()
        exit_code = safe_shell_exec.execute(lscpu_cmd, stdout=output, stderr=output)
        if exit_code != 0:
            raise RuntimeError("{cmd} failed with exit code {exit_code}".format(
                cmd=lscpu_cmd, exit_code=exit_code))
        return int(yaml.safe_load(output.getvalue())[LSFUtils._THREAD_KEY])



# Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import time

from horovod.runner.util.threads import in_thread
from horovod.spark.task import task_info, task_service
from horovod.spark.task.task_info import get_available_devices
from horovod.spark.driver import driver_service
from horovod.runner.common.util import codec, secret


def _parent_process_monitor(initial_ppid):
    try:
        while True:
            if initial_ppid != os.getppid():
                # Parent process died, terminate
                os._exit(1)
            time.sleep(1)
    except:
        # Avoids an error message during Python interpreter shutdown.
        pass


def task_exec(driver_addresses, settings, rank_env, local_rank_env):
    # Die if parent process terminates
    in_thread(target=_parent_process_monitor, args=(os.getppid(),))

    key = codec.loads_base64(os.environ[secret.HOROVOD_SECRET_KEY])
    rank = int(os.environ[rank_env])
    local_rank = int(os.environ[local_rank_env])
    driver_client = driver_service.SparkDriverClient(driver_addresses, key,
                                                     verbose=settings.verbose)

    # tell driver about local rank and rank
    # in elastic mode the driver already knows this mapping
    # for simplicity we keep code paths the same for elastic and static mode
    host_hash = os.environ['HOROVOD_HOSTNAME']
    task_index = driver_client.set_local_rank_to_rank(host_hash, local_rank, rank)

    # gather available resources from task service
    task_addresses = driver_client.all_task_addresses(task_index)
    task_client = task_service.SparkTaskClient(task_index, task_addresses, key,
                                               verbose=settings.verbose)
    task_info.set_resources(task_client.resources())

    fn, args, kwargs = driver_client.code()
    result = fn(*args, **kwargs)
    task_client.register_code_result(result)



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
# Modifications copyright (C) 2019 Uber Technologies, Inc.
# Modifications copyright Microsoft
# Modifications copyright (C) 2020, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""Inter-process communication using MPI."""

import re
import tensorflow as tf
from tensorflow.python.framework import load_library
from tensorflow.python.framework import ops
from tensorflow.python.platform import resource_loader

from horovod.common.util import check_installed_version, get_ext_suffix, \
    get_average_backwards_compatibility_fun, gpu_available, num_rank_is_power_2
from horovod.common.basics import HorovodBasics as _HorovodBasics
from horovod.common.process_sets import _setup as _setup_process_sets
from horovod.common.process_sets import ProcessSet, global_process_set, add_process_set, remove_process_set, \
    _temp_process_set_object
from horovod.tensorflow.util import _executing_eagerly


def _load_library(name):
    """Loads a .so file containing the specified operators.

    Args:
      name: The name of the .so file to load.

    Raises:
      NotFoundError if were not able to load .so file.
    """
    filename = resource_loader.get_path_to_datafile(name)
    library = load_library.load_op_library(filename)
    return library

# Check possible symbol not found error from tensorflow version mismatch
try:
    MPI_LIB = _load_library('mpi_lib' + get_ext_suffix())
except Exception as e:
    check_installed_version('tensorflow', tf.__version__, e)
    raise e
else:
    check_installed_version('tensorflow', tf.__version__)

_basics = _HorovodBasics(__file__, 'mpi_lib')

# import basic methods
shutdown = _basics.shutdown
is_initialized = _basics.is_initialized
start_timeline = _basics.start_timeline
stop_timeline = _basics.stop_timeline
size = _basics.size
local_size = _basics.local_size
cross_size = _basics.cross_size
rank = _basics.rank
local_rank = _basics.local_rank
cross_rank = _basics.cross_rank
mpi_threads_supported = _basics.mpi_threads_supported
mpi_enabled = _basics.mpi_enabled
mpi_built = _basics.mpi_built
gloo_enabled = _basics.gloo_enabled
gloo_built = _basics.gloo_built
nccl_built = _basics.nccl_built
ddl_built = _basics.ddl_built
ccl_built = _basics.ccl_built
cuda_built = _basics.cuda_built
rocm_built = _basics.rocm_built

# import reduction op values
Average = _basics.Average
Sum = _basics.Sum
Adasum = _basics.Adasum
Min = _basics.Min
Max = _basics.Max
Product = _basics.Product

def init(*args, **kwargs):
    _basics.init(*args, **kwargs)
    # Call set up again to make sure the basics is in sync
    _setup_process_sets(_basics)

is_homogeneous = _basics.is_homogeneous

handle_average_backwards_compatibility = get_average_backwards_compatibility_fun(_basics)

check_num_rank_power_of_2 = num_rank_is_power_2

_setup_process_sets(_basics)


# This function will create a default device map which includes all visible devices.
# Please run this function in a subprocess
def _check_has_gpu():
    import tensorflow as tf
    return tf.test.is_gpu_available()


def _normalize_name(name):
    """Normalizes operation name to TensorFlow rules."""
    return re.sub('[^a-zA-Z0-9_]', '_', name)


def _allreduce(tensor, name=None, op=Sum, prescale_factor=1.0, postscale_factor=1.0,
               ignore_name_scope=False, process_set=global_process_set):
    """An op which reduces an input tensor over all the Horovod processes. The
    default reduction is a sum.

    The reduction operation is keyed by the name of the op. The tensor type and
    shape must be the same on all Horovod processes for a given name. The reduction
    will not start until all processes are ready to send and receive the tensor.

    Returns:
      A tensor of the same shape and type as `tensor`, summed across all
      processes.
    """
    if name is None and not _executing_eagerly():
        name = 'HorovodAllreduce_%s' % _normalize_name(tensor.name)
    return MPI_LIB.horovod_allreduce(tensor, name=name, reduce_op=op,
                                     prescale_factor=prescale_factor,
                                     postscale_factor=postscale_factor,
                                     ignore_name_scope=ignore_name_scope,
                                     process_set_id=process_set.process_set_id)


@ops.RegisterGradient('HorovodAllreduce')
def _allreduce_grad(op, grad):
    """Gradient for allreduce op.

    Args:
      op: An operation.
      grad: `Tensor` gradient with respect to the output of the op.

    Returns:
      The gradient with respect to the input of the op.
    """
    reduce_op = op.get_attr('reduce_op')
    prescale_factor = op.get_attr('prescale_factor')
    postscale_factor = op.get_attr('postscale_factor')
    ignore_name_scope = op.get_attr('ignore_name_scope')
    process_set_id = op.get_attr('process_set_id')
    return _allreduce(grad, op=reduce_op, prescale_factor=prescale_factor,
                      postscale_factor=postscale_factor,
                      ignore_name_scope=ignore_name_scope,
                      process_set=_temp_process_set_object(process_set_id))


def _grouped_allreduce(tensors, name=None, op=Sum, prescale_factor=1.0, postscale_factor=1.0,
                       ignore_name_scope=False, process_set=global_process_set):
    """An op which reduces input tensors over all the Horovod processes. The
    default reduction is a sum.

    The reduction operations are keyed by the name of the op. Reductions are
    performed across tensors in the same list position. The tensor type and
    shape must be the same on all Horovod processes for tensors sharing
    positions in the input tensor list. The reduction will not start until all
    processes are ready to send and receive the tensors.

    Returns:
      A list of tensors of the same shape and type as those in `tensors`,
      summed across all processes.
    """
    if name is None and not _executing_eagerly():
        name = _normalize_name('HorovodGroupedAllreduce_%s_%s' % (tensors[0].name, tensors[-1].name))
    return MPI_LIB.horovod_grouped_allreduce(tensors, name=name, reduce_op=op,
                                             prescale_factor=prescale_factor,
                                             postscale_factor=postscale_factor,
                                             ignore_name_scope=ignore_name_scope,
                                             process_set_id=process_set.process_set_id)


@ops.RegisterGradient('HorovodGroupedAllreduce')
def _grouped_allreduce_grad(op, *grads):
    """Gradient for the grouped allreduce op.

    Args:
      op: An operation.
      grads: List of `Tensor` gradients with respect to the outputs of the op.

    Returns:
      The gradients with respect to the inputs of the op.
    """
    reduce_op = op.get_attr('reduce_op')
    prescale_factor = op.get_attr('prescale_factor')
    postscale_factor = op.get_attr('postscale_factor')
    ignore_name_scope = op.get_attr('ignore_name_scope')
    process_set_id = op.get_attr('process_set_id')
    # TODO(joshr): should this be done as separate allreduce ops?
    return _grouped_allreduce(list(grads), op=reduce_op, prescale_factor=prescale_factor,
                              postscale_factor=postscale_factor,
                              ignore_name_scope=ignore_name_scope,
                              process_set=_temp_process_set_object(process_set_id))


def allgather(tensor, name=None, ignore_name_scope=False, process_set=global_process_set):
    """An op which concatenates the input tensor with the same input tensor on
    all other Horovod processes.

    The concatenation is done on the first dimension, so the input tensors on the
    different processes must have the same rank and shape, except for the first
    dimension, which is allowed to be different.

    Returns:
      A tensor of the same type as `tensor`, concatenated on dimension zero
      across all processes. The shape is identical to the input shape, except for
      the first dimension, which may be greater and is the sum of all first
      dimensions of the tensors in different Horovod processes.
    """
    if name is None and not _executing_eagerly():
        name = 'HorovodAllgather_%s' % _normalize_name(tensor.name)
    return MPI_LIB.horovod_allgather(tensor, name=name,
                                     ignore_name_scope=ignore_name_scope,
                                     process_set_id=process_set.process_set_id)


@ops.RegisterGradient('HorovodAllgather')
def _allgather_grad(op, grad):
    """Gradient for allgather op.

    Args:
      op: An operation.
      grad: `Tensor` gradient with respect to the output of the op.

    Returns:
      The gradient with respect to the input of the op.
    """
    ignore_name_scope = op.get_attr('ignore_name_scope')
    process_set_id = op.get_attr('process_set_id')
    temp_process_set_object = _temp_process_set_object(process_set_id)
    grad = _allreduce(grad, op=Average, ignore_name_scope=ignore_name_scope,
                      process_set=temp_process_set_object)

    with tf.device('/cpu:0'):
        # Keep the tensor of split sizes on CPU.
        x = op.inputs[0]
        d = tf.shape(x)
        d = tf.reshape(d[0], [1])

        s = temp_process_set_object.size()
        d = tf.reshape(
            allgather(d, ignore_name_scope=ignore_name_scope, process_set=temp_process_set_object),
            [s])

    splits = tf.split(grad, num_or_size_splits=d, axis=0)
    return splits[temp_process_set_object.rank()]


def grouped_allgather(tensors, name=None, ignore_name_scope=False, process_set=global_process_set):
    """An op which concatenates the input tensor with the same input tensor on
    all other Horovod processes.

    The concatenation is done on the first dimension, so the input tensors on the
    different processes must have the same rank and shape, except for the first
    dimension, which is allowed to be different.

    Returns:
      A list of tensors of the same rank and type as `tensor`, concatenated on
      dimension zero across all processes. For each returned tensor the shape
      is identical to the input shape, except for the first dimension, which
      may be greater and is the sum of all first dimensions of the tensors
      in different Horovod processes.
    """
    if name is None and not _executing_eagerly():
        name = _normalize_name('HorovodGroupedAllgather_%s_%s' % (tensors[0].name, tensors[-1].name))
    return MPI_LIB.horovod_grouped_allgather(tensors, name=name,
                                             ignore_name_scope=ignore_name_scope,
                                             process_set_id=process_set.process_set_id)


@ops.RegisterGradient('HorovodGroupedAllgather')
def _grouped_allgather_grad(op, *grads):
    """Gradient for the grouped allgather op.

    Args:
      op: An operation.
      grad: List of `Tensor` gradients with respect to the outputs of the op.

    Returns:
      The gradients with respect to the inputs of the op.
    """
    ignore_name_scope = op.get_attr('ignore_name_scope')
    process_set_id = op.get_attr('process_set_id')
    temp_process_set_object = _temp_process_set_object(process_set_id)
    # Not using _grouped_allreduce here because all its input tensors need to be of the same dtype, but TensorFlow may
    # give us float32 zero gradients here despite non-zero gradients being, e.g, float64.
    grads = [_allreduce(g, op=Average, ignore_name_scope=ignore_name_scope, process_set=temp_process_set_object)
             for g in grads]

    split_sizes = []
    with tf.device('/cpu:0'):
        # Keep the split size tensors on CPU.
        for x in op.inputs:
            d = tf.shape(x)
            d = tf.reshape(d[0], [1])

            s = temp_process_set_object.size()
            d = tf.reshape(
                allgather(d, ignore_name_scope=ignore_name_scope, process_set=temp_process_set_object),
                [s])
            split_sizes.append(d)

    splits_list = [tf.split(grad, num_or_size_splits=d, axis=0)
                   for grad, d in zip(grads, split_sizes)]
    return [splits[temp_process_set_object.rank()] for splits in splits_list]


def broadcast(tensor, root_rank, name=None, ignore_name_scope=False, process_set=global_process_set):
    """An op which broadcasts the input tensor on root rank to the same input tensor
    on all other Horovod processes.

    The broadcast operation is keyed by the name of the op. The tensor type and
    shape must be the same on all Horovod processes for a given name. The broadcast
    will not start until all processes are ready to send and receive the tensor.

    Returns:
      A tensor of the same shape and type as `tensor`, with the value broadcasted
      from root rank.
    """
    if name is None and not _executing_eagerly():
        name = 'HorovodBroadcast_%s' % _normalize_name(tensor.name)
    return MPI_LIB.horovod_broadcast(tensor, name=name, root_rank=root_rank,
                                     ignore_name_scope=ignore_name_scope,
                                     process_set_id=process_set.process_set_id)


@ops.RegisterGradient('HorovodBroadcast')
def _broadcast_grad(op, grad):
    """Gradient for broadcast op.

    Args:
      op: An operation.
      grad: `Tensor` gradient with respect to the output of the op.

    Returns:
      The gradient with respect to the input of the op.
    """
    root_rank = op.get_attr('root_rank')
    ignore_name_scope = op.get_attr('ignore_name_scope')
    process_set_id = op.get_attr('process_set_id')
    grad_reduced = _allreduce(grad, op=Average,
                              ignore_name_scope=ignore_name_scope,
                              process_set=_temp_process_set_object(process_set_id))
    if rank() != root_rank:
        return grad_reduced * 0
    return grad_reduced


def broadcast_(variables, root_rank, name=None, process_set=global_process_set):
    """An op which broadcasts the input variables from the root rank to the same
    input variables on all other Horovod processes. The operation is performed
    in-place.

    The broadcast operation is keyed by the name of the op combined with the names
    of the variables. The variable type and shape must be the same on all Horovod
    processes for any given name. The broadcast will not start until all processes
    are ready to send and receive all variables. In each process all variables need
    to be located on the same device (CPU or GPU).

    Note: This is only supported with TensorFlow 2.6 or later.

    Returns:
      The tensor values of the updated `variables` as broadcasted from root rank.
    """
    from packaging import version
    if version.parse(tf.__version__) < version.parse('2.6.0'):
        raise NotImplementedError("In-place broadcasts are only supported with TensorFlow 2.6 or later")

    from tensorflow.python.ops import resource_variable_ops
    if all(resource_variable_ops.is_resource_variable(var) for var in variables):
        with tf.control_dependencies(
            [MPI_LIB.horovod_broadcast_inplace_resource([var.handle for var in variables],
                                                        variable_names=[var.name for var in variables],
                                                        name=name,
                                                        root_rank=root_rank,
                                                        process_set_id=process_set.process_set_id)]):
            return [var.read_value() for var in variables]
    elif all(not resource_variable_ops.is_resource_variable(var) for var in variables):
        return MPI_LIB.horovod_broadcast_inplace(variables, variable_names=[var.name for var in variables],
                                                 name=name, root_rank=root_rank,
                                                 process_set_id=process_set.process_set_id)
    else:
        raise ValueError("All variables passed to broadcast_() should be of the same kind, resource or reference")


def alltoall(tensor, splits=None, name=None, ignore_name_scope=False, process_set=global_process_set):
    """An op that scatters slices of the input tensor to all other Horovod processes
    and returns a tensor of gathered slices from all other Horovod processes.

    The slicing is done on the first dimension, so the input tensors on the
    different processes must have the same rank and shape, except for the first
    dimension, which is allowed to be different.

    Arguments:
        tensor: A tensor to distribute with alltoall.
        splits: A tensor of integers in rank order describing how many
                elements in `tensor` to send to each worker.  Splitting is
                applied along the first dimension of `tensor`. If `splits` is
                not provided, the first dimension is split equally by the
                number of Horovod processes.
        name: A name of the alltoall operation.
        ignore_name_scope: If True, ignores any outer name scope applied by
                           TensorFlow in the name used by the Horovod operation.

    Returns:
      1) A tensor of the same type as `tensor`, concatenated on dimension zero
         across all processes. The shape is identical to the input shape, except for
         the first dimension, which may be greater and is the sum of all first
         dimensions of the gathered tensor slices from different Horovod processes.
      2) If `splits` has been provided: A tensor of integers in rank order
         describing how many elements in the output tensor have been received
         from each worker.
    """
    # If splits not provided, create empty tensor as placeholder
    splits_ = tf.convert_to_tensor(splits) if splits is not None else tf.constant([], dtype=tf.int32)

    if name is None and not _executing_eagerly():
        name = 'HorovodAlltoall_%s' % _normalize_name(tensor.name)
    output, rsplits = MPI_LIB.horovod_alltoall(tensor, splits=splits_, name=name,
                                               ignore_name_scope=ignore_name_scope,
                                               process_set_id=process_set.process_set_id)
    return (output, rsplits) if splits is not None else output

@ops.RegisterGradient('HorovodAlltoall')
def _alltoall_grad(op, grad_wrt_output, grad_wrt_received_splits):
    """Gradient for alltoall op.

    Args:
      op: Original operation.
      grad_wrt_output: `Tensor` gradient with respect to the output of the op.
      grad_wrt_received_splits: dead argument (integer output)

    Returns:
      The gradient with respect to the input of the op.
    """
    ignore_name_scope = op.get_attr('ignore_name_scope')
    process_set_id = op.get_attr('process_set_id')
    recvsplits = op.outputs[1]

    grad_wrt_tensor, _ = alltoall(grad_wrt_output, splits=recvsplits, ignore_name_scope=ignore_name_scope,
                                  process_set=_temp_process_set_object(process_set_id))
    grad_wrt_splits = None # not differentiable (integer variable)

    return [grad_wrt_tensor, grad_wrt_splits]


def _reducescatter(tensor, name=None, op=Sum, ignore_name_scope=False,
                   process_set=global_process_set, prescale_factor=1.0,
                   postscale_factor=1.0):
    """An op which reduces an input tensor over all the Horovod processes, then
    scatters the result across all the Horovod processes. The default reduction
    is a sum.

    The reduction operation is keyed by the name of the op. The tensor type and
    shape must be the same on all Horovod processes for a given name. The reduction
    will not start until all processes are ready to send and receive the tensor.

    Returns:
      A tensor of the same rank and type as `tensor`. The shape is identical to the
        input shape, except for the first dimension, which will be divided across
        the different Horovod processes.
    """
    if name is None and not _executing_eagerly():
        name = 'HorovodReducescatter_%s' % _normalize_name(tensor.name)
    return MPI_LIB.horovod_reducescatter(tensor, name=name, reduce_op=op,
                                         ignore_name_scope=ignore_name_scope,
                                         process_set_id=process_set.process_set_id,
                                         prescale_factor=prescale_factor,
                                         postscale_factor=postscale_factor)


@ops.RegisterGradient('HorovodReducescatter')
def _reducescatter_grad(op, grad):
    """Gradient for reducescatter op.

    Args:
      op: An operation.
      grad: `Tensor` gradient with respect to the output of the op.

    Returns:
      The gradient with respect to the input of the op.
    """
    ignore_name_scope = op.get_attr('ignore_name_scope')
    process_set_id = op.get_attr('process_set_id')
    reduce_op = op.get_attr('reduce_op')
    prescale_factor = op.get_attr('prescale_factor')
    postscale_factor = op.get_attr('postscale_factor')
    process_set = _temp_process_set_object(process_set_id)
    if reduce_op == Sum:
        grad *= process_set.size()
    if prescale_factor != 1.0:
        grad *= prescale_factor
    if postscale_factor != 1.0:
        grad *= postscale_factor
    return allgather(grad, ignore_name_scope=ignore_name_scope,
                     process_set=process_set)


def _grouped_reducescatter(tensors, name=None, op=Sum, ignore_name_scope=False,
                           process_set=global_process_set,
                           prescale_factor=1.0, postscale_factor=1.0):
    """An op which sums an input tensor over all the Horovod processes, then
    scatters the result across all the Horovod processes.

    The reduction operations are keyed by the name of the op. Reductions are
    performed across tensors in the same list position. The tensor type and
    shape must be the same on all Horovod processes for tensors sharing
    positions in the input tensor list. The reduction will not start until all
    processes are ready to send and receive the tensors.

    Returns:
      A list of tensors of the same rank and type as those in `tensors`.
      For each returned tensor the shape is identical to the corresponding
      input shape, except for the first dimension, which will be divided across
      the different Horovod processes.
    """
    if name is None and not _executing_eagerly():
        name = _normalize_name('HorovodGroupedReducescatter_%s_%s' % (tensors[0].name, tensors[-1].name))
    return MPI_LIB.horovod_grouped_reducescatter(tensors, name=name, reduce_op=op,
                                                 ignore_name_scope=ignore_name_scope,
                                                 process_set_id=process_set.process_set_id,
                                                 prescale_factor=prescale_factor,
                                                 postscale_factor=postscale_factor)


@ops.RegisterGradient('HorovodGroupedReducescatter')
def _grouped_reducescatter_grad(op, *grads):
    """Gradient for the grouped reducescatter op.

    Args:
      op: An operation.
      grads: List of `Tensor` gradients with respect to the outputs of the op.

    Returns:
      The gradients with respect to the inputs of the op.
    """
    ignore_name_scope = op.get_attr('ignore_name_scope')
    process_set_id = op.get_attr('process_set_id')
    reduce_op = op.get_attr('reduce_op')
    prescale_factor = op.get_attr('prescale_factor')
    postscale_factor = op.get_attr('postscale_factor')
    process_set = _temp_process_set_object(process_set_id)
    if reduce_op == Sum:
        grads = [grad * process_set.size() for grad in grads]
    if prescale_factor != 1.0:
        grads = [grad * prescale_factor for grad in grads]
    if postscale_factor != 1.0:
        grads = [grad * postscale_factor for grad in grads]
    # Not using grouped_allgather here because all its input tensors need to be of the same dtype, but TensorFlow may
    # give us float32 zero gradients here despite non-zero gradients being, e.g, float64.
    return [allgather(g, ignore_name_scope=ignore_name_scope, process_set=process_set) for g in grads]


def join():
    """An op to indicate that the rank finished processing data.

    All ranks that did not call join() continue to process allreduce
    operations. This op is not done before all ranks have joined.

    Returns:
      An integer scalar containing the the last rank that joined.
    """
    return MPI_LIB.horovod_join()


def size_op(process_set_id=0, name=None):
    """An op that returns the number of Horovod processes.

    This operation determines the return value at the graph execution time,
    rather than at the graph construction time, and so allows for a graph to be
    constructed in a different environment than where it will be executed.

    Returns:
      An integer scalar containing the number of Horovod processes.
    """
    return MPI_LIB.horovod_size(process_set_id=process_set_id, name=name)


ops.NotDifferentiable('HorovodSize')


def process_set_included_op(process_set_id=0, name=None):
    """An op that 0 or 1 depending on whether the current process is
    included in the specified process set or an error code:
    HOROVOD_PROCESS_SET_ERROR_INIT if Horovod is not initialized,
    HOROVOD_PROCESS_SET_ERROR_UNKNOWN_SET if the process set is unknown.

    This operation determines the return value at the graph execution time,
    rather than at the graph construction time, and so allows for a graph to be
    constructed in a different environment than where it will be executed.

    Returns:
      An integer scalar with value 0, 1, or an error code.
    """
    return MPI_LIB.horovod_process_set_included(process_set_id=process_set_id, name=name)


ops.NotDifferentiable('HorovodProcessSetIncluded')


def local_size_op(name=None):
    """An op that returns the number of Horovod processes within the
    node the current process is running on.

    This operation determines the return value at the graph execution time,
    rather than at the graph construction time, and so allows for a graph to be
    constructed in a different environment than where it will be executed.

    Returns:
      An integer scalar containing the number of local Horovod processes.
    """
    return MPI_LIB.horovod_local_size(name=name)


ops.NotDifferentiable('HorovodLocalSize')


def rank_op(name=None):
    """An op that returns the Horovod rank of the calling process.

    This operation determines the return value at the graph execution time,
    rather than at the graph construction time, and so allows for a graph to be
    constructed in a different environment than where it will be executed.

    Returns:
      An integer scalar with the Horovod rank of the calling process.
    """
    return MPI_LIB.horovod_rank(name=name)


ops.NotDifferentiable('HorovodRank')


def local_rank_op(name=None):
    """An op that returns the local Horovod rank of the calling process, within the
    node that it is running on. For example, if there are seven processes running
    on a node, their local ranks will be zero through six, inclusive.

    This operation determines the return value at the graph execution time,
    rather than at the graph construction time, and so allows for a graph to be
    constructed in a different environment than where it will be executed.

    Returns:
      An integer scalar with the local Horovod rank of the calling process.
    """
    return MPI_LIB.horovod_rank(name=name)


ops.NotDifferentiable('HorovodLocalRank')



# Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
# Modifications copyright Microsoft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


class HorovodInternalError(RuntimeError):
    """Internal error raised when a Horovod collective operation (e.g., allreduce) fails.

    This is handled in elastic mode as a recoverable error, and will result in a reset event.
    """
    pass


class HostsUpdatedInterrupt(RuntimeError):
    """Internal interrupt event indicating that the set of hosts in the job has changed.

    In elastic mode, this will result in a reset event without a restore to committed state.
    """
    def __init__(self, skip_sync):
        self.skip_sync = skip_sync


def get_version_mismatch_message(name, version, installed_version):
    return f'Framework {name} installed with version {installed_version} but found version {version}.\n\
             This can result in unexpected behavior including runtime errors.\n\
             Reinstall Horovod using `pip install --no-cache-dir` to build with the new version.'


class HorovodVersionMismatchError(ImportError):
    """Internal error raised when the runtime version of a framework mismatches its version at
    Horovod installation time.
    """
    def __init__(self, name, version, installed_version):
        super().__init__(get_version_mismatch_message(name, version, installed_version))
        self.name = name
        self.version = version
        self.installed_version = installed_version



# Copyright 2020 Uber Technologies, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import io
import sys
import traceback

from horovod.runner.common.util import safe_shell_exec


def execute(command, env=None):
    """
    Executes the command and returns stdout and stderr as a string, together with the exit code.
    :param command: command to execute
    :param env: environment variables to use
    :return: (output, exit code) or None on failure
    """
    output = io.StringIO()
    try:
        exit_code = safe_shell_exec.execute(command, env=env, stdout=output, stderr=output)
        output_msg = output.getvalue()
    except Exception:
        print(traceback.format_exc(), file=sys.stderr)
        return None
    finally:
        output.close()

    return output_msg, exit_code



# Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from queue import Queue, Empty
from threading import Thread, Event


class BaseDataLoader(object):
    def __len__(self):
        """
        Length of the batches to be loaded.
        """
        raise NotImplementedError()

    def _iterate(self):
        """
        Interface for the implimentation of iterate batches
        """
        raise NotImplementedError()

    def __iter__(self):
        """
        Starting iteration and get batchs
        """
        for batch in self._iterate():
            yield self._process_batch(batch)

    def _process_batch(self, batch):
        """
        Hook to modify batch before output. Will be override by trainer to reshape the data
        as needed. Please do not override it.
        """
        return batch


class AsyncDataLoaderMixin(object):
    """
    Async Mixin on top of implementation of BaseDataLoader. It contains a seperate thread
    which reads batch from self._iterate() and push them in the queue. The self.__iter__() function
    will pop the batch from the queue.
    If async_loader_queue_size is set to 0, the data loader will not work in async mode.
    For example:
        class PytorchAsyncDataLoader(AsyncDataLoaderMixin, PytorchDataLoader):
    """

    def __init__(self, async_loader_queue_size=64, debug_data_loader=False, *args, **kwargs):
        """
        initialize the async data loader. Need to add this in the __init__() of the implementation
        """
        self.async_loader_queue_size = async_loader_queue_size
        self.debug_data_loader = debug_data_loader
        super().__init__(*args, **kwargs)

        print(f"Apply the AsyncDataLoaderMixin on top of the data loader, async_loader_queue_size={async_loader_queue_size}. ")

        if self.async_loader_queue_size > 0:
            self.finished_event = Event()
            self.queue = Queue(self.async_loader_queue_size)
            self.thread = Thread(target=self._async_worker)
            self.thread.daemon = True
            self.started = False

    def close_async_loader(self):
        """
        Close the async data loader.
        """
        print(f"close_async_loader[{self.async_loader_queue_size}], Closing the AsyncDataLoaderMixin.")
        if self.async_loader_queue_size > 0 and self.started:
            self.finished_event.set()
            c = 0
            while True:
                try:
                    # Drain buffer
                    self.queue.get_nowait()
                    if self.debug_data_loader:
                        print(f"close_async_loader[{self.async_loader_queue_size}], discarded batch #{c} from Queue.")

                    c += 1

                    # Force break out if hanging. We assume hanging if already pop items more than twice of the size of the queue.
                    if c > max(2 * self.async_loader_queue_size, 200):
                        print(f"close_async_loader: ERROR!!! Force break out after {c} times get_nowait.")
                        break
                except Empty:
                    break
            if self.debug_data_loader:
                print(f"close_async_loader[{self.async_loader_queue_size}], joining...")

            self.thread.join()

        print(f"close_async_loader[{self.async_loader_queue_size}] is closed.")

    def _async_worker(self):
        """
        Start worker thread to load data asynchronously.
        User need to implement self._iterate() to read the data.
        """
        try:
            c = 0
            while not self.finished_event.is_set():
                for batch in self._iterate():
                    if self.finished_event.is_set():
                        break
                    self.queue.put(batch)

                    if self.debug_data_loader:
                        print(f"_async_worker[{self.async_loader_queue_size}], push batch #{c}.")
                        c += 1

                if self.debug_data_loader:
                    print(f"_async_worker[{self.async_loader_queue_size}], finish reading at #{c}, reset debugging counter, append None to queue.")
                    c = 0

                self.queue.put(None)
        except Exception as ex:
            self.queue.put(ex)
            self.queue.put(None)
        finally:
            self.queue.put(None)

        print(f"_async_worker[{self.async_loader_queue_size}], stoped")

    def __iter__(self):
        """
        Override the __iter__() to iterate data asynchronously to produce batchs.
        Will procude batchs from the queue which were generated by self._iterate().
        """

        print("Start generating batches from async data loader.")
        if self.async_loader_queue_size > 0:
            if not self.started:
                self.started = True
                self.thread.start()

            c = 0
            while True:
                batch = self.queue.get()

                if self.debug_data_loader:
                    print(f"__iter__[{self.async_loader_queue_size}], get batch #{c}.")
                    c += 1

                if batch is None:
                    if self.debug_data_loader:
                        print(f"__iter__[{self.async_loader_queue_size}], get None from queue at #{c}.")
                    break
                if isinstance(batch, Exception):
                    raise batch

                yield self._process_batch(batch)
        else:
            for batch in self._iterate():
                yield self._process_batch(batch)



"""Module for supporting multi-node callbacks.

This module maintains a set of global variables for each
Ray process.
"""

import warnings

_queue = None
_queue_set = False
warning_raised = False


def configure(queue):
    """Sets the proper global variables."""

    global _queue
    global _queue_set
    _queue = queue
    # We maintain a _queue_set variable because somehow
    # the _queue variable makes a lot of unnecessary remote calls.
    _queue_set = True


def log(info_dict):
    """Sends info dict back to the driver, where it will be processed.

    Args:
        info_dict (dict): Dictionary of serializable elements.
    """
    global warning_raised
    global _queue_set

    if _queue_set:
        _queue.put(info_dict)
    elif not warning_raised:
        warnings.warn("`ray_logger.log` called but logging is not configured.")
        warning_raised = True



# Copyright 2022 G-Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import argparse
import logging
import signal
import sys

from pyspark.sql import SparkSession

from horovod.runner.common.service.compute_service import ComputeService
from horovod.runner.common.util import secret
from horovod.spark import run
from horovod.tensorflow.data.compute_service import TfDataServiceConfig, compute_worker_fn

if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument("configfile", type=str,
                        help=f"The path to store the compute service config file.")

    parser.add_argument("--dispatchers", required=False, default=1, type=int,
                        help=f"The number of dispatcher to support.",
                        dest="dispatchers")

    parser.add_argument("--dispatcher-side", required=False, default='compute', type=str,
                        help=f"Where do the dispatcher run? On 'compute' side or 'training' side.",
                        dest="dispatcher_side")

    parsed_args = parser.parse_args()

    spark = SparkSession.builder.getOrCreate()
    spark_context = spark.sparkContext
    workers = spark_context.defaultParallelism

    if workers % parsed_args.dispatchers:
        raise ValueError(f'Number of processes ({workers}) must be '
                         f'a multiple of number of dispatchers ({parsed_args.dispatchers}).')
    workers_per_dispatcher = workers // parsed_args.dispatchers

    key = secret.make_secret_key()
    compute = ComputeService(parsed_args.dispatchers, workers_per_dispatcher, key=key)

    compute_config = TfDataServiceConfig(
        dispatchers=parsed_args.dispatchers,
        workers_per_dispatcher=workers_per_dispatcher,
        dispatcher_side=parsed_args.dispatcher_side,
        addresses=compute.addresses(),
        key=key
    )
    compute_config.write(parsed_args.configfile)

    def _exit_gracefully():
        logging.info('Spark driver receiving SIGTERM. Exiting gracefully')
        spark_context.stop()

    signal.signal(signal.SIGTERM, _exit_gracefully)

    ret = run(compute_worker_fn,
              args=(compute_config,),
              stdout=sys.stdout,
              stderr=sys.stderr,
              num_proc=workers,
              verbose=2)

    compute.shutdown()
    spark.stop()

    sys.exit(ret)



# Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


class TaskInfo(object):
    def __init__(self):
        self.resources = dict()


_info = TaskInfo()


def get_available_devices():
    if 'gpu' not in _info.resources:
        return []
    return _info.resources['gpu'].addresses


def set_resources(resources):
    _info.resources = resources



# Copyright 2020 Google Research. All Rights Reserved.
# Modifications copyright (c) 2020, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import tensorflow as tf
from horovod.tensorflow.mpi_ops import _allreduce
from horovod.tensorflow.mpi_ops import size, rank
from horovod.tensorflow.mpi_ops import Sum

class SyncBatchNormalization(tf.keras.layers.BatchNormalization):
  """ Synchronous batch normalization. Stats are synchronized across all workers during training. """

  def __init__(self, fused=False, **kwargs):
    if fused in (True, None):
      raise ValueError('SyncBatchNormalization does not support fused=True.')
    if not kwargs.get('name', None):
      kwargs['name'] = 'sync_batch_normalization'
    super(SyncBatchNormalization, self).__init__(fused=fused, **kwargs)

  def _moments(self, inputs, reduction_axes, keep_dims, **kwargs):
    """Compute the mean and variance: it overrides the original _moments."""
    worker_mean, worker_variance = super(SyncBatchNormalization, self)._moments(
      inputs, reduction_axes, keep_dims=keep_dims, **kwargs)

    if size() > 1:
      # Compute variance using: Var[X] = E[X^2] - E[X]^2.
      worker_square_of_mean = tf.math.square(worker_mean)
      worker_mean_of_square = worker_variance + worker_square_of_mean

      # Average stats across all workers
      worker_stack = tf.stack([worker_mean, worker_mean_of_square])
      group_stack = _allreduce(worker_stack, op=Sum)
      group_stack /= size()
      group_mean, group_mean_of_square = tf.unstack(group_stack)

      group_variance = group_mean_of_square - tf.math.square(group_mean)

      return (group_mean, group_variance)
    else:
      return (worker_mean, worker_variance)

  def call(self, *args, **kwargs):
    outputs = super(SyncBatchNormalization, self).call(*args, **kwargs)
    try:
      # A temporary hack for TF1 compatibility with Keras batch norm.
      # Ops are added to tf.GraphKeys.UPDATE_OPS manually to mimic
      # behavior of TF1 batch norm layer for use in control dependencies.
      for u in self.updates:
        tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, u)
    except AttributeError:
      pass
    return outputs



# Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
# Modifications copyright Microsoft
# Modifications copyright (C) 2020, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Load all the necessary PyTorch C types.
import torch

import warnings

from horovod.common.basics import HorovodBasics as _HorovodBasics
from horovod.common.exceptions import HorovodInternalError
from horovod.common.process_sets import _setup as _setup_process_sets
from horovod.common.process_sets import ProcessSet, global_process_set, add_process_set, remove_process_set
from horovod.common.util import check_installed_version, get_average_backwards_compatibility_fun, gpu_available, num_rank_is_power_2

from horovod.torch.compression import Compression

# Check possible symbol not found error from pytorch version mismatch
try:
    from horovod.torch import mpi_lib_v2 as mpi_lib
except Exception as e:
    check_installed_version('pytorch', torch.__version__, e)
    raise e
else:
    check_installed_version('pytorch', torch.__version__)

_NULL = ""

_basics = _HorovodBasics(__file__, 'mpi_lib_v2')

# import basic methods
is_initialized = _basics.is_initialized
start_timeline = _basics.start_timeline
stop_timeline = _basics.stop_timeline
size = _basics.size
local_size = _basics.local_size
cross_size = _basics.cross_size
rank = _basics.rank
local_rank = _basics.local_rank
cross_rank = _basics.cross_rank
mpi_threads_supported = _basics.mpi_threads_supported
mpi_enabled = _basics.mpi_enabled
mpi_built = _basics.mpi_built
gloo_enabled = _basics.gloo_enabled
gloo_built = _basics.gloo_built
nccl_built = _basics.nccl_built
ddl_built = _basics.ddl_built
ccl_built = _basics.ccl_built
cuda_built = _basics.cuda_built
rocm_built = _basics.rocm_built

def shutdown(*args, **kwargs):
    mpi_lib.horovod_torch_reset()
    return _basics.shutdown(*args, **kwargs)

def init(*args, **kwargs):
    global _handle_map
    _handle_map = {}
    _basics.init(*args, **kwargs)
    # Call set up again to make sure the basics is in sync
    _setup_process_sets(_basics)

# import reduction op values
Average = _basics.Average
Sum = _basics.Sum
Adasum = _basics.Adasum
Min = _basics.Min
Max = _basics.Max
Product = _basics.Product

is_homogeneous = _basics.is_homogeneous

handle_average_backwards_compatibility = get_average_backwards_compatibility_fun(_basics)

_setup_process_sets(_basics)


# Schema: handle -> input, output
# We keep input in order to make sure it does not get garbage collected
# before the operation is finished.
_handle_map = {}


def _check_function(function_factory, tensor):
    function = function_factory(tensor)
    if not hasattr(mpi_lib, function):
        raise ValueError('Tensor type %s is not supported.' % tensor.type())
    if not tensor.is_contiguous():
        raise ValueError('Tensor is required to be contiguous.')
    return function


def _allreduce_function_factory(tensor):
    return 'horovod_torch_allreduce_async_' + tensor.type().replace('.', '_')


def _allreduce_async(tensor, output, name, op, prescale_factor, postscale_factor, process_set: ProcessSet):
    # Set the divisor for reduced gradients to average when necessary
    if op == Average:
        if rocm_built():
            # For ROCm, perform averaging at framework level
            divisor = process_set.size()
            op = Sum
        else:
            divisor = 1

    elif op == Adasum:
        if process_set != global_process_set:
            raise NotImplementedError("Adasum does not support non-global process sets yet.")
        if tensor.device.type != 'cpu' and gpu_available('torch'):
            if nccl_built():
                if not is_homogeneous():
                    raise NotImplementedError('Running GPU Adasum on heterogeneous cluster is not supported yet.')
                elif not num_rank_is_power_2(int(size() / local_size())):
                    raise NotImplementedError('Running GPU Adasum with non-power of 2 nodes is not supported yet.')
                if rocm_built():
                    # For ROCm, perform averaging at framework level
                    divisor = local_size()
                else:
                    divisor = 1
            else:
                warnings.warn('Adasum reduction does not currently support GPU reduction using MPI. Tensors are '
                              'copied to CPU memory instead. To use Adasum for GPU reduction, please compile Horovod '
                              'with HOROVOD_GPU_OPERATIONS=NCCL.')
                divisor = 1
        else:
            if not num_rank_is_power_2(size()):
                raise NotImplementedError('Running Adasum with non-power of 2 ranks is not supported yet.')
            divisor = 1
    else:
        divisor = 1

    function = _check_function(_allreduce_function_factory, tensor)
    try:
        handle = getattr(mpi_lib, function)(tensor, output, divisor,
                                            name.encode() if name is not None else _NULL, op,
                                            prescale_factor, postscale_factor, process_set.process_set_id)
    except RuntimeError as e:
        raise HorovodInternalError(e)
    _handle_map[handle] = (tensor, output)
    return handle


def allreduce_async(tensor, average=None, name=None, op=None,
                    prescale_factor=1.0, postscale_factor=1.0,
                    process_set=global_process_set):
    """
    A function that performs asynchronous averaging or summation of the input tensor
    over all the Horovod processes. The input tensor is not modified.

    The reduction operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The reduction will not start until all processes
    are ready to send and receive the tensor.

    Arguments:
        tensor: A tensor to reduce.
        average:
            .. warning:: .. deprecated:: 0.19.0

                Use `op` instead. Will be removed in v1.0.

        name: A name of the reduction operation.
        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        prescale_factor: Multiplicative factor to scale tensor before allreduce.
        postscale_factor: Multiplicative factor to scale tensor after allreduce.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A handle to the allreduce operation that can be used with `poll()` or
        `synchronize()`.
    """
    op = handle_average_backwards_compatibility(op, average)
    output = tensor.new(tensor.shape)
    return _allreduce_async(tensor, output, name, op, prescale_factor, postscale_factor, process_set)


class HorovodAllreduce(torch.autograd.Function):
    """An autograd function that performs allreduce on a tensor."""

    @staticmethod
    def forward(ctx, tensor, average, name, op, prescale_factor, postscale_factor, process_set):
        ctx.average = average
        ctx.op = op
        ctx.prescale_factor = prescale_factor
        ctx.postscale_factor = postscale_factor
        ctx.process_set = process_set
        handle = allreduce_async(tensor, average, name, op, prescale_factor, postscale_factor, process_set)
        return synchronize(handle)

    @staticmethod
    def backward(ctx, grad_output):
        return allreduce(grad_output, average=ctx.average, op=ctx.op,
                         prescale_factor=ctx.prescale_factor,
                         postscale_factor=ctx.postscale_factor,
                         process_set=ctx.process_set), None, None, None, None, None, None


def allreduce(tensor, average=None, name=None, compression=Compression.none, op=None,
              prescale_factor=1.0, postscale_factor=1.0, process_set=global_process_set):
    """
    A function that performs averaging or summation of the input tensor over all the
    Horovod processes. The input tensor is not modified.

    The reduction operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The reduction will not start until all processes
    are ready to send and receive the tensor.

    This acts as a thin wrapper around an autograd function.  If your input
    tensor requires gradients, then callings this function will allow gradients
    to be computed and backpropagated.

    Arguments:
        tensor: A tensor to reduce.
        average:
            .. warning:: .. deprecated:: 0.19.0

                Use `op` instead. Will be removed in v1.0.

        name: A name of the reduction operation.
        compression: Compression algorithm used during allreduce to reduce the amount
                     of data sent during the each parameter update step.  Defaults to
                     not using compression.
        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        prescale_factor: Multiplicative factor to scale tensor before allreduce.
        postscale_factor: Multiplicative factor to scale tensor after allreduce.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A tensor of the same shape and type as `tensor`, averaged or summed across all
        processes.
    """
    tensor_compressed, ctx = compression.compress(tensor)
    summed_tensor_compressed = HorovodAllreduce.apply(tensor_compressed, average, name, op,
                                                      prescale_factor, postscale_factor,
                                                      process_set)
    return compression.decompress(summed_tensor_compressed, ctx)


def allreduce_async_(tensor, average=None, name=None, op=None,
                     prescale_factor=1.0, postscale_factor=1.0,
                     process_set=global_process_set):
    """
    A function that performs asynchronous in-place averaging or summation of the input
    tensor over all the Horovod processes.

    The reduction operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The reduction will not start until all processes
    are ready to send and receive the tensor.

    Arguments:
        tensor: A tensor to reduce.
        average:
            .. warning:: .. deprecated:: 0.19.0

                Use `op` instead. Will be removed in v1.0.

        name: A name of the reduction operation.
        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        prescale_factor: Multiplicative factor to scale tensor before allreduce.
        postscale_factor: Multiplicative factor to scale tensor after allreduce.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A handle to the allreduce operation that can be used with `poll()` or
        `synchronize()`.
    """
    op = handle_average_backwards_compatibility(op, average)
    return _allreduce_async(tensor, tensor, name, op, prescale_factor, postscale_factor, process_set)


def allreduce_(tensor, average=None, name=None, op=None,
               prescale_factor=1.0, postscale_factor=1.0,
               process_set=global_process_set):
    """
    A function that performs in-place averaging or summation of the input tensor over
    all the Horovod processes.

    The reduction operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The reduction will not start until all processes
    are ready to send and receive the tensor.

    Arguments:
        tensor: A tensor to reduce.
        average:
            .. warning:: .. deprecated:: 0.19.0

                Use `op` instead. Will be removed in v1.0.

        name: A name of the reduction operation.
        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        prescale_factor: Multiplicative factor to scale tensor before allreduce.
        postscale_factor: Multiplicative factor to scale tensor after allreduce.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A tensor of the same shape and type as `tensor`, averaged or summed across all
        processes.
    """
    handle = allreduce_async_(tensor, average, name, op, prescale_factor, postscale_factor, process_set)
    return synchronize(handle)


def _grouped_allreduce_function_factory(tensor):
    return 'horovod_torch_grouped_allreduce_async_' + tensor.type().replace('.', '_')


def _grouped_allreduce_async(tensors, outputs, name, op, prescale_factor, postscale_factor, process_set: ProcessSet):
    # Set the divisor for reduced gradients to average when necessary
    if op == Average:
        if rocm_built():
            # For ROCm, perform averaging at framework level
            divisor = process_set.size()
            op = Sum
        else:
            divisor = 1
    elif op == Adasum:
        if process_set != global_process_set:
            raise NotImplementedError("Adasum does not support non-global process sets yet.")
        if tensors[0].device.type != 'cpu' and gpu_available('torch'):
            if nccl_built():
                if not is_homogeneous():
                    raise NotImplementedError('Running GPU Adasum on heterogeneous cluster is not supported yet.')
                elif not num_rank_is_power_2(int(size() / local_size())):
                    raise NotImplementedError('Running GPU Adasum with non-power of 2 nodes is not supported yet.')
                if rocm_built():
                    # For ROCm, perform averaging at framework level
                    divisor = local_size()
                else:
                    divisor = 1
            else:
                warnings.warn('Adasum reduction does not currently support GPU reduction using MPI. Tensors are '
                              'copied to CPU memory instead. To use Adasum for GPU reduction, please compile Horovod '
                              'with HOROVOD_GPU_OPERATIONS=NCCL.')
                divisor = 1
        else:
            if not num_rank_is_power_2(size()):
                raise NotImplementedError('Running Adasum with non-power of 2 ranks is not supported yet.')
            divisor = 1
    else:
        divisor = 1

    function = _check_function(_grouped_allreduce_function_factory, tensors[0])
    try:
        handle = getattr(mpi_lib, function)(tensors, outputs, divisor,
                                            name.encode() if name is not None else _NULL, op,
                                            prescale_factor, postscale_factor, process_set.process_set_id)
    except RuntimeError as e:
        raise HorovodInternalError(e)
    _handle_map[handle] = (tuple(tensors), tuple(outputs))
    return handle


def grouped_allreduce_async(tensors, average=None, name=None, op=None,
                            prescale_factor=1.0, postscale_factor=1.0,
                            process_set=global_process_set):
    """
    A function that performs asynchronous averaging or summation of the input tensor
    list over all the Horovod processes. The input tensors are not modified.

    The reduction operations are keyed by the base name. If a base name is not
    provided, an incremented auto-generated base name is used. Reductions are
    performed across tensors in the same list position. The tensor type and
    shape must be the same on all Horovod processes for tensors sharing
    positions in the input tensor list. The reduction will not start until all
    processes are ready to send and receive the tensors.

    Arguments:
        tensors: A list of tensors to reduce.
        average:
            .. warning:: .. deprecated:: 0.19.0

                Use `op` instead. Will be removed in v1.0.

        name: A base name to use for the group reduction operation.
        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        prescale_factor: Multiplicative factor to scale tensor before allreduce.
        postscale_factor: Multiplicative factor to scale tensor after allreduce.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A handle to the group allreduce operation that can be used with `poll()` or
        `synchronize()`.
    """
    op = handle_average_backwards_compatibility(op, average)
    outputs = [t.new(t.shape) for t in tensors]
    return _grouped_allreduce_async(tensors, outputs, name, op, prescale_factor, postscale_factor, process_set)


class HorovodGroupedAllreduce(torch.autograd.Function):
    """An autograd function that performs allreduce on a list of tensors."""

    @staticmethod
    def forward(ctx, average, name, op, prescale_factor, postscale_factor, process_set: ProcessSet, *tensors):
        ctx.average = average
        ctx.op = op
        ctx.prescale_factor = prescale_factor
        ctx.postscale_factor = postscale_factor
        ctx.process_set = process_set
        handle = grouped_allreduce_async(list(tensors), average, name, op, prescale_factor, postscale_factor,
                                         process_set)
        return synchronize(handle)

    @staticmethod
    def backward(ctx, *grad_output):
        grad_reduced = grouped_allreduce(list(grad_output), average=ctx.average, op=ctx.op,
                                         prescale_factor=ctx.prescale_factor,
                                         postscale_factor=ctx.postscale_factor,
                                         process_set=ctx.process_set)
        return (None, None, None, None, None, None, *grad_reduced)


def grouped_allreduce(tensors, average=None, name=None, compression=Compression.none, op=None,
                      prescale_factor=1.0, postscale_factor=1.0, process_set=global_process_set):
    """
    A function that performs averaging or summation of the input tensor
    list over all the Horovod processes. The input tensors are not modified.

    The reduction operations are keyed by the base name. If a base name is not
    provided, an incremented auto-generated base name is used. Reductions are
    performed across tensors in the same list position. The tensor type and
    shape must be the same on all Horovod processes for tensors sharing
    positions in the input tensor list. The reduction will not start until all
    processes are ready to send and receive the tensors.

    This acts as a thin wrapper around an autograd function.  If your input
    tensors require gradients, then calling this function will allow gradients
    to be computed and backpropagated.

    Arguments:
        tensors: A list of tensors to reduce.
        average:
            .. warning:: .. deprecated:: 0.19.0

                Use `op` instead. Will be removed in v1.0.

        name: A base name to use for the group reduction operation.
        compression: Compression algorithm used during allreduce to reduce the amount
                     of data sent during the each parameter update step.  Defaults to
                     not using compression.
        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        prescale_factor: Multiplicative factor to scale tensor before allreduce.
        postscale_factor: Multiplicative factor to scale tensor after allreduce.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A list containing tensors of the same shape and type as in `tensors`,
        averaged or summed across all processes.
    """
    tensors_compressed, ctxs = zip(*[compression.compress(t) for t in tensors])
    summed_tensors_compressed = HorovodGroupedAllreduce.apply(average, name, op,
                                                              prescale_factor, postscale_factor,
                                                              process_set, *tensors_compressed)
    return [compression.decompress(t, ctx) for t, ctx in zip(summed_tensors_compressed, ctxs)]


def grouped_allreduce_async_(tensors, average=None, name=None, op=None,
                             prescale_factor=1.0, postscale_factor=1.0,
                             process_set=global_process_set):
    """
    A function that performs asynchronous in-place averaging or summation of the input
    tensors over all the Horovod processes.

    The reduction operations are keyed by the base name. If a base name is not
    provided, an incremented auto-generated base name is used. Reductions are
    performed across tensors in the same list position. The tensor type and
    shape must be the same on all Horovod processes for tensors sharing
    positions in the input tensor list. The reduction will not start until all
    processes are ready to send and receive the tensors.

    Arguments:
        tensors: A list of tensors to reduce.
        average:
            .. warning:: .. deprecated:: 0.19.0

                Use `op` instead. Will be removed in v1.0.

        name: A base name to use for the group reduction operation.
        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        prescale_factor: Multiplicative factor to scale tensor before allreduce.
        postscale_factor: Multiplicative factor to scale tensor after allreduce.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A handle to the group allreduce operation that can be used with `poll()` or
        `synchronize()`.
    """
    op = handle_average_backwards_compatibility(op, average)
    return _grouped_allreduce_async(tensors, tensors, name, op, prescale_factor, postscale_factor, process_set)


def grouped_allreduce_(tensors, average=None, name=None, op=None,
                       prescale_factor=1.0, postscale_factor=1.0,
                       process_set=global_process_set):
    """
    A function that performs in-place averaging or summation of the input tensors over
    all the Horovod processes.

    The reduction operations are keyed by the base name. If a base name is not
    provided, an incremented auto-generated base name is used. Reductions are
    performed across tensors in the same list position. The tensor type and
    shape must be the same on all Horovod processes for tensors sharing
    positions in the input tensor list. The reduction will not start until all
    processes are ready to send and receive the tensors.

    Arguments:
        tensors: A list of tensors to reduce.
        average:
            .. warning:: .. deprecated:: 0.19.0

                Use `op` instead. Will be removed in v1.0.

        name: A base name to use for the group reduction operation.
        op: The reduction operation to combine tensors across different ranks.
            Supported op values are Sum, Average, Min, Max, and Product. Defaults
            to Average if None is given.
        prescale_factor: Multiplicative factor to scale tensor before allreduce.
        postscale_factor: Multiplicative factor to scale tensor after allreduce.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A list containing tensors of the same shape and type as in `tensors`,
        averaged or summed across all processes.
    """
    handle = grouped_allreduce_async_(tensors, average, name, op, prescale_factor, postscale_factor, process_set)
    return synchronize(handle)


def sparse_allreduce_async(tensor, name, op, process_set=global_process_set):
    # Allgather aggregates along the first dimension, so we need to transpose the
    # indices to enforce correct concatenation behavior, then transpose back prior to
    # constructing the new aggregated sparse gradient
    t = tensor
    indices_handle = allgather_async(t._indices().transpose(0, 1).contiguous(), name=f'{name}.indices',
                                     process_set=process_set)
    values_handle = allgather_async(t._values(), name=f'{name}.values', process_set=process_set)

    def handle():
        # We need to sync values handle firstly for torch nightly >= 10.0
        # Issue: https://github.com/horovod/horovod/issues/2961
        values = synchronize(values_handle)
        indices = synchronize(indices_handle)

        values = (values / process_set.size()) if op == Average else values

        if indices.dim() == 0 or values.dim() == 0:
            return t.new().resize_as_(t)
        return t.new(indices.transpose(0, 1), values, t.size())

    return handle


def _allgather_function_factory(tensor):
    return 'horovod_torch_allgather_async_' + tensor.type().replace('.', '_')


def _allgather_async(tensor, output, name, process_set: ProcessSet):
    function = _check_function(_allgather_function_factory, tensor)
    try:
        handle = getattr(mpi_lib, function)(
            tensor, output, name.encode() if name is not None else _NULL,
            process_set.process_set_id)
    except RuntimeError as e:
        raise HorovodInternalError(e)
    _handle_map[handle] = (tensor, output)
    return handle


def allgather_async(tensor, name=None, process_set=global_process_set):
    """
    A function that asynchronously concatenates the input tensor with the same input
    tensor on all other Horovod processes. The input tensor is not modified.

    The concatenation is done on the first dimension, so the input tensors on the
    different processes must have the same rank and shape, except for the first
    dimension, which is allowed to be different.

    Arguments:
        tensor: A tensor to allgather.
        name: A name of the allgather operation.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A handle to the allgather operation that can be used with `poll()` or
        `synchronize()`.
    """
    output = tensor.new()
    return _allgather_async(tensor, output, name, process_set)


class HorovodAllgather(torch.autograd.Function):
    """An autograd function that performs allgather on a tensor."""

    @staticmethod
    def forward(ctx, tensor, name, process_set: ProcessSet):
        ctx.dim = tensor.shape[0]
        ctx.process_set = process_set
        handle = allgather_async(tensor, name, process_set)
        return synchronize(handle)

    @staticmethod
    def backward(ctx, grad_output):
        grad_reduced = allreduce(grad_output, average=True, process_set=ctx.process_set)

        dim_t = torch.IntTensor([ctx.dim])
        dim = allgather(dim_t, process_set=ctx.process_set).view(ctx.process_set.size())

        r = ctx.process_set.rank()
        offset = torch.sum(dim.narrow(0, 0, r)).item() if r != 0 else 0
        return grad_reduced.narrow(0, offset, ctx.dim), None, None


def allgather(tensor, name=None, process_set=global_process_set):
    """
    A function that concatenates the input tensor with the same input tensor on
    all other Horovod processes. The input tensor is not modified.

    The concatenation is done on the first dimension, so the corresponding
    input tensors on the different processes must have the same rank and
    shape, except for the first dimension, which is allowed to be different.

    This acts as a thin wrapper around an autograd function.  If your input
    tensor requires gradients, then callings this function will allow gradients
    to be computed and backpropagated.

    Arguments:
        tensor: A tensor to allgather.
        name: A name of the allgather operation.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A tensor of the same type as `tensor`, concatenated on dimension zero
        across all processes. The shape is identical to the input shape, except for
        the first dimension, which may be greater and is the sum of all first
        dimensions of the tensors in different Horovod processes.
    """
    return HorovodAllgather.apply(tensor, name, process_set)


def _grouped_allgather_function_factory(tensor):
    return 'horovod_torch_grouped_allgather_async_' + tensor.type().replace('.', '_')


def _grouped_allgather_async(tensors, outputs, name, process_set: ProcessSet):
    function = _check_function(_grouped_allgather_function_factory, tensors[0])
    try:
        handle = getattr(mpi_lib, function)(
            tensors, outputs, name.encode() if name is not None else _NULL,
            process_set.process_set_id)
    except RuntimeError as e:
        raise HorovodInternalError(e)
    _handle_map[handle] = (tuple(tensors), tuple(outputs))
    return handle


def grouped_allgather_async(tensors, name=None, process_set=global_process_set):
    """
    A function that asynchronously concatenates each input tensor with the corresponding
    input tensor on all other Horovod processes for a list of input tensors. The input
    tensors are not modified.

    The concatenation is done on the first dimension, so the input tensors on the
    different processes must have the same rank and shape, except for the first
    dimension, which is allowed to be different.

    Arguments:
        tensors: A list of tensors to allgather.
        name: A base name to use for the group allgather operation.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A handle to the group allgather operation that can be used with `poll()` or
        `synchronize()`.
    """
    outputs = [t.new() for t in tensors]
    return _grouped_allgather_async(tensors, outputs, name, process_set)


class HorovodGroupedAllgather(torch.autograd.Function):
    """An autograd function that performs allgather on a list of tensors."""

    @staticmethod
    def forward(ctx, name, process_set: ProcessSet, *tensors):
        ctx.dims = [t.shape[0] for t in tensors]
        ctx.process_set = process_set
        handle = grouped_allgather_async(list(tensors), name, process_set)
        return synchronize(handle)

    @staticmethod
    def backward(ctx, *grad_output):
        grad_reduced = grouped_allreduce(list(grad_output), average=True, process_set=ctx.process_set)

        dim_ts = [torch.IntTensor([dim]) for dim in ctx.dims]
        dims = [dt.view(ctx.process_set.size())
                for dt in grouped_allgather(dim_ts, process_set=ctx.process_set)]

        r = ctx.process_set.rank()
        offsets = [torch.sum(dim.narrow(0, 0, r)).item() if r != 0 else 0
                   for dim in dims]
        result = [gr.narrow(0, offset, dim)
                  for gr, offset, dim in zip(grad_reduced, offsets, ctx.dims)]
        return (None, None, *result)


def grouped_allgather(tensors, name=None, process_set=global_process_set):
    """
    A function that concatenates each input tensor with the corresponding
    input tensor on all other Horovod processes for a list of input tensors.
    The input tensors are not modified.

    The concatenation is done on the first dimension, so the corresponding input
    tensors on the different processes must have the same rank and shape, except
    for the first dimension, which is allowed to be different.

    Arguments:
        tensors: A list of tensors to allgather.
        name: A base name to use for the group allgather operation.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A list containing tensors of the same type as in `tensors`. Each tensor
        is concatenated on dimension zero across all processes. Its shape is
        identical to the corresponding input shape, expect for the first
        dimension, which may be greater and is the sum of all first dimensions
        of the corresponding tensor in different Horovod processes.
    """
    return HorovodGroupedAllgather.apply(name, process_set, *tensors)


def _broadcast_function_factory(tensor):
    return 'horovod_torch_broadcast_async_' + tensor.type().replace('.', '_')


def _broadcast_async(tensor, output, root_rank, name, process_set: ProcessSet):
    function = _check_function(_broadcast_function_factory, tensor)
    try:
        handle = getattr(mpi_lib, function)(
            tensor, output, root_rank, name.encode() if name is not None else _NULL,
            process_set.process_set_id)
    except RuntimeError as e:
        raise HorovodInternalError(e)
    _handle_map[handle] = (tensor, output)
    return handle


def broadcast_async(tensor, root_rank, name=None, process_set=global_process_set):
    """
    A function that asynchronously broadcasts the input tensor on root rank to the same
    input tensor on all other Horovod processes. The input tensor is not modified.

    The broadcast operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The broadcast will not start until all processes
    are ready to send and receive the tensor.

    Arguments:
        tensor: A tensor to broadcast.
        root_rank: The rank to broadcast the value from.
        name: A name of the broadcast operation.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A handle to the broadcast operation that can be used with `poll()` or
        `synchronize()`.
    """
    output = tensor.new(tensor.shape)
    return _broadcast_async(tensor, output, root_rank, name, process_set)


class HorovodBroadcast(torch.autograd.Function):
    """An autograd function that broadcasts a tensor."""

    @staticmethod
    def forward(ctx, tensor, root_rank, name, process_set: ProcessSet):
        ctx.root_rank = root_rank
        ctx.process_set = process_set
        handle = broadcast_async(tensor, root_rank, name, process_set)
        return synchronize(handle)

    @staticmethod
    def backward(ctx, grad_output):
        grad_reduced = allreduce(grad_output, average=True, process_set=ctx.process_set)
        if rank() != ctx.root_rank:
            grad_reduced *= 0
        return grad_reduced, None, None, None


def broadcast(tensor, root_rank, name=None, process_set=global_process_set):
    """
    A function that broadcasts the input tensor on root rank to the same input tensor
    on all other Horovod processes. The input tensor is not modified.

    The broadcast operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The broadcast will not start until all processes
    are ready to send and receive the tensor.

    This acts as a thin wrapper around an autograd function.  If your input
    tensor requires gradients, then callings this function will allow gradients
    to be computed and backpropagated.

    Arguments:
        tensor: A tensor to broadcast.
        root_rank: The rank to broadcast the value from.
        name: A name of the broadcast operation.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A tensor of the same shape and type as `tensor`, with the value broadcasted
        from root rank.
    """
    return HorovodBroadcast.apply(tensor, root_rank, name, process_set)


def broadcast_async_(tensor, root_rank, name=None, process_set=global_process_set):
    """
    A function that asynchronously broadcasts the input tensor on root rank to the same
    input tensor on all other Horovod processes. The operation is performed in-place.

    The broadcast operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The broadcast will not start until all processes
    are ready to send and receive the tensor.

    Arguments:
        tensor: A tensor to broadcast.
        root_rank: The rank to broadcast the value from.
        name: A name of the broadcast operation.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A handle to the broadcast operation that can be used with `poll()` or
        `synchronize()`.
    """
    return _broadcast_async(tensor, tensor, root_rank, name, process_set)


def broadcast_(tensor, root_rank, name=None, process_set=global_process_set):
    """
    A function that broadcasts the input tensor on root rank to the same input tensor
    on all other Horovod processes. The operation is performed in-place.

    The broadcast operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The broadcast will not start until all processes
    are ready to send and receive the tensor.

    Arguments:
        tensor: A tensor to broadcast.
        root_rank: The rank to broadcast the value from.
        name: A name of the broadcast operation.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        A tensor of the same shape and type as `tensor`, with the value broadcasted
        from root rank.
    """
    handle = broadcast_async_(tensor, root_rank, name, process_set)
    return synchronize(handle)

def _alltoall_function_factory(tensor):
    return 'horovod_torch_alltoall_async_' + tensor.type().replace('.', '_')

def _alltoall_async(tensor, splits, output, output_received_splits, name, process_set: ProcessSet):
    if splits is None:
        # If splits not provided, create empty tensor as placeholder
        splits = torch.tensor([], dtype=torch.int32, device='cpu')
    elif not isinstance(splits, torch.Tensor):
        splits = torch.tensor(splits, dtype=torch.int32, device='cpu')
    function = _check_function(_alltoall_function_factory, tensor)
    try:
        handle = getattr(mpi_lib, function)(
            tensor, splits, output, output_received_splits, name.encode() if name is not None else _NULL,
            process_set.process_set_id)
    except RuntimeError as e:
        raise HorovodInternalError(e)
    _handle_map[handle] = (tensor, splits, (output, output_received_splits))
    return handle


def alltoall_async(tensor, splits=None, name=None, process_set=global_process_set):
    """
    A function that scatters slices of the input tensor to all other Horovod processes
    and returns a tensor of gathered slices from all other Horovod processes. The input
    tensor is not modified.

    The slicing is done on the first dimension, so the input tensors on
    the different processes must have the same rank and shape, except for the
    first dimension, which is allowed to be different.

    Arguments:
        tensor: A tensor to distribute with alltoall.
        splits: A tensor of integers in rank order describing how many
                elements in `tensor` to send to each worker.  Splitting is
                applied along the first dimension of `tensor`. If `splits` is
                not provided, the first dimension is split equally by the
                number of Horovod processes.
        name: A name of the alltoall operation.
        process_set: Process set object to limit this operation to a subset of
                Horovod processes. Default is the global process set.

    Returns:
        A handle to the alltoall operation that can be used with `poll()` or
        `synchronize()`.
    """
    output = tensor.new()
    if isinstance(splits, torch.Tensor):
        output_received_splits = splits.new()
    else:
        output_received_splits = torch.empty(size(), dtype=torch.int32, device='cpu')
    return _alltoall_async(tensor, splits, output, output_received_splits, name, process_set)


class HorovodAlltoall(torch.autograd.Function):
    """An autograd function that performs alltoall on a tensor."""

    @staticmethod
    def forward(ctx, tensor, splits, name, process_set: ProcessSet):
        handle = alltoall_async(tensor, splits, name, process_set)
        output, received_splits = synchronize(handle)

        ctx.process_set = process_set
        ctx.recvsplits = received_splits
        if splits is None:
            return output
        else:
            ctx.mark_non_differentiable(received_splits)
            return output, received_splits

    @staticmethod
    def backward(ctx, grad_output, *dead_gradients):
        grad_wrt_tensor, _ = alltoall(grad_output, splits=ctx.recvsplits,
                                      process_set=ctx.process_set)
        return grad_wrt_tensor, None, None, None


def alltoall(tensor, splits=None, name=None, process_set=global_process_set):
    """
    A function that scatters slices of the input tensor to all other Horovod processes
    and returns a tensor of gathered slices from all other Horovod processes. The input
    tensor is not modified.

    The slicing is done on the first dimension, so the input tensors on
    the different processes must have the same rank and shape, except for the
    first dimension, which is allowed to be different.

    This acts as a thin wrapper around an autograd function.  If your input
    tensor requires gradients, then callings this function will allow gradients
    to be computed and backpropagated.

    Arguments:
        tensor: A tensor to distribute with alltoall.
        splits: A tensor of integers in rank order describing how many
                elements in `tensor` to send to each worker.  Splitting is
                applied along the first dimension of `tensor`. If `splits` is
                not provided, the first dimension is split equally by the
                number of Horovod processes.
        name: A name of the alltoall operation.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.

    Returns:
        1) A tensor containing the gathered tensor data from all workers.
        2) If `splits` has been provided: A tensor of integers in rank order
           describing how many elements in the output tensor have been received
           from each worker.
     """
    return HorovodAlltoall.apply(tensor, splits, name, process_set)


def _reducescatter_function_factory(tensor):
    return 'horovod_torch_reducescatter_async_' + tensor.type().replace('.', '_')


def _reducescatter_async(tensor, output, name, op, process_set: ProcessSet,
                         prescale_factor, postscale_factor):
    function = _check_function(_reducescatter_function_factory, tensor)
    try:
        handle = getattr(mpi_lib, function)(tensor, output,
                                            name.encode() if name is not None else _NULL,
                                            op, process_set.process_set_id,
                                            prescale_factor, postscale_factor)
    except RuntimeError as e:
        raise HorovodInternalError(e)
    _handle_map[handle] = (tensor, output)
    return handle


def reducescatter_async(tensor, name=None, op=Average, process_set=global_process_set,
                        prescale_factor=1.0, postscale_factor=1.0):
    """
    A function that performs asynchronous reduction of the input tensor over all the
    Horovod processes, then scatters the results across all Horovod processes. The
    input tensor is not modified.

    The reduction operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The reduction will not start until all processes
    are ready to send and receive the tensor.

    The input tensors on the different processes must have the same rank and shape. The
    output tensor will be the same rank on all processes, but the first dimension may
    be different.

    Arguments:
        tensor: A tensor to average and sum.
        name: A name of the reduction operation.
        op: The reduction operation to combine tensors across different ranks.
            Defaults to Average.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.
        prescale_factor: Multiplicative factor to scale tensor before reducescatter.
        postscale_factor: Multiplicative factor to scale tensor after reducescatter.

    Returns:
        A handle to the reducescatter operation that can be used with `poll()` or
        `synchronize()`.
    """
    output = tensor.new()
    return _reducescatter_async(tensor, output, name, op, process_set,
                                prescale_factor, postscale_factor)


class HorovodReducescatter(torch.autograd.Function):
    """An autograd function that performs reducescatter on a tensor."""

    @staticmethod
    def forward(ctx, tensor, name, op, process_set, prescale_factor, postscale_factor):
        ctx.op = op
        ctx.process_set = process_set
        ctx.prescale_factor = prescale_factor
        ctx.postscale_factor = postscale_factor
        handle = reducescatter_async(tensor, name, op, process_set, prescale_factor, postscale_factor)
        return synchronize(handle)

    @staticmethod
    def backward(ctx, grad_output):
        if ctx.op == Sum:
            grad_output *= ctx.process_set.size()
        if ctx.prescale_factor != 1.0:
            grad_output *= ctx.prescale_factor
        if ctx.postscale_factor != 1.0:
            grad_output *= ctx.postscale_factor

        return allgather(grad_output, process_set=ctx.process_set), None, None, None, None, None


def reducescatter(tensor, name=None, compression=Compression.none, op=Average,
                  process_set=global_process_set, prescale_factor=1.0, postscale_factor=1.0):
    """
    A function that performs reduction of the input tensor over all the Horovod
    processes, then scatters the results across all Horovod processes. The input tensor
    is not modified.

    The reduction operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The reduction will not start until all processes
    are ready to send and receive the tensor.

    Arguments:
        tensor: A tensor to average/sum and scatter.
        name: A name of the reduction operation.
        compression: Compression algorithm used during reducescatter to reduce the amount
                     of data sent during the each parameter update step.  Defaults to
                     not using compression.
        op: The reduction operation to combine tensors across different ranks.
            Defaults to Average.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.
        prescale_factor: Multiplicative factor to scale tensor before reducescatter.
        postscale_factor: Multiplicative factor to scale tensor after reducescatter.

    Returns:
        A tensor of the same rank and type as `tensor` across all processes. The shape
        is identical to the input shape, except for the first dimension, which will be
        divided across the different Horovod processes.
    """
    tensor_compressed, ctx = compression.compress(tensor)
    reduced_tensor_compressed = HorovodReducescatter.apply(tensor_compressed, name, op, process_set,
                                                           prescale_factor, postscale_factor)
    return compression.decompress(reduced_tensor_compressed, ctx)


def _grouped_reducescatter_function_factory(tensor):
    return 'horovod_torch_grouped_reducescatter_async_' + tensor.type().replace('.', '_')


def _grouped_reducescatter_async(tensors, outputs, name, op, process_set: ProcessSet,
                                 prescale_factor, postscale_factor):
    function = _check_function(_grouped_reducescatter_function_factory, tensors[0])
    try:
        handle = getattr(mpi_lib, function)(tensors, outputs,
                                            name.encode() if name is not None else _NULL,
                                            op, process_set.process_set_id,
                                            prescale_factor, postscale_factor)
    except RuntimeError as e:
        raise HorovodInternalError(e)
    _handle_map[handle] = (tuple(tensors), tuple(outputs))
    return handle


def grouped_reducescatter_async(tensors, name=None, op=Average, process_set=global_process_set,
                                prescale_factor=1.0, postscale_factor=1.0):
    """
    A function that performs asynchronous reduction of a list of input tensors over all the
    Horovod processes, then scatters the results across all Horovod processes. The
    input tensors are not modified.

    The reduction operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. For each of the input tensors, the type and shape must
    be the same on all Horovod processes for a given name. The reduction will not start
    until all processes are ready to send and receive the tensors.

    The input tensor at some place in the list tensor must have the same rank and shape
    on the different processes. The corresponding output tensor will be the same rank on
    all processes, but the first dimension may be different.

    Arguments:
        tensors: A list of tensors to average and sum.
        name: A base name to use for the group reduction operation.
        op: The reduction operation to combine tensors across different ranks.
            Defaults to Average.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.
        prescale_factor: Multiplicative factor to scale tensors before reducescatter.
        postscale_factor: Multiplicative factor to scale tensors after reducescatter.

    Returns:
        A handle to the group reducescatter operation that can be used with `poll()` or
        `synchronize()`.
    """
    outputs = [t.new() for t in tensors]
    return _grouped_reducescatter_async(tensors, outputs, name, op, process_set,
                                        prescale_factor, postscale_factor)


class HorovodGroupedReducescatter(torch.autograd.Function):
    """An autograd function that performs reducescatter on a list of tensors."""

    @staticmethod
    def forward(ctx, name, op, process_set, prescale_factor, postscale_factor, *tensors):
        ctx.op = op
        ctx.process_set = process_set
        ctx.prescale_factor = prescale_factor
        ctx.postscale_factor = postscale_factor
        handle = grouped_reducescatter_async(list(tensors), name, op, process_set,
                                             prescale_factor, postscale_factor, )
        return synchronize(handle)

    @staticmethod
    def backward(ctx, *grad_output):
        if ctx.op == Sum:
            grad_output = [g * ctx.process_set.size() for g in grad_output]
        if ctx.prescale_factor != 1.0:
            grad_output = [ctx.prescale_factor * g for g in grad_output]
        if ctx.postscale_factor != 1.0:
            grad_output = [ctx.postscale_factor * g for g in grad_output]

        return (None, None, None, None, None,
                *grouped_allgather(grad_output, process_set=ctx.process_set))


def grouped_reducescatter(tensors, name=None, compression=Compression.none, op=Average,
                          process_set=global_process_set, prescale_factor=1.0, postscale_factor=1.0):
    """
    A function that performs reduction of a list of input tensors over all the
    Horovod processes, then scatters the results across all Horovod processes. The
    input tensors are not modified.

    The reduction operation is keyed by the name. If name is not provided, an incremented
    auto-generated name is used. The tensor type and shape must be the same on all
    Horovod processes for a given name. The reduction will not start until all processes
    are ready to send and receive the tensor.

    Arguments:
        tensors: A list of tensors to average and sum.
        name: A base name to use for the group reduction operation.
        compression: Compression algorithm used during reducescatter to reduce the amount
                     of data sent during the each parameter update step.  Defaults to
                     not using compression.
        op: The reduction operation to combine tensors across different ranks.
            Defaults to Average.
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.
        prescale_factor: Multiplicative factor to scale tensors before reducescatter.
        postscale_factor: Multiplicative factor to scale tensors after reducescatter.


    Returns:
        A list containing tensors of the same rank and type as in `tensors`. For each
        tensor the shape is identical to the input shape, except for the first dimension,
        which will be divided across the different Horovod processes.
    """
    tensors_compressed = []
    ctxs = []
    for tensor in tensors:
        tensor_compressed, ctx = compression.compress(tensor)
        tensors_compressed.append(tensor_compressed)
        ctxs.append(ctx)
    reduced_tensors_compressed = HorovodGroupedReducescatter.apply(name, op, process_set,
                                                                   prescale_factor, postscale_factor,
                                                                   *tensors_compressed)
    return [compression.decompress(reduced_tensor_compressed, ctx)
            for reduced_tensor_compressed, ctx in zip(reduced_tensors_compressed, ctxs)]


def poll(handle):
    """
    Polls an allreduce, allgather, alltoall, broadcast, or reducescatter handle to determine whether
    underlying asynchronous operation has completed. After `poll()` returns `True`,
    `synchronize()` will return without blocking.

    Arguments:
        handle: A handle returned by an allreduce, allgather, alltoall, broadcast, or reducescatter
                asynchronous operation.

    Returns:
        A flag indicating whether the operation has completed.
    """
    return mpi_lib.horovod_torch_poll(handle) != 0


def synchronize(handle):
    """
    Synchronizes an asynchronous allreduce, allgather, alltoall, broadcast, or reducescatter operation
    until  it's completed. Returns the result of the operation.

    Arguments:
        handle: A handle returned by an allreduce, allgather, alltoall, broadcast, or reducescatter
                asynchronous operation.

    Returns:
        A single output tensor of the operation or a tuple of multiple output tensors.
    """
    if handle not in _handle_map:
        return

    try:
        mpi_lib.horovod_torch_wait_and_clear(handle)
        output = _handle_map.pop(handle)[-1]
        return output
    except RuntimeError as e:
        _handle_map.pop(handle, None)
        raise HorovodInternalError(e)


def join(device=-1) -> int:
    """A function that indicates that the rank finished processing data.

    All ranks that did not call join() continue to process allreduce operations.
    This function blocks Python thread until all ranks join.

    Arguments:
        device: An id of the device to create temprorary zero tensors (default -1, CPU)

    Returns:
        Id of the rank that joined last.
    """
    output = torch.tensor(-1, dtype=torch.int, device=torch.device("cpu"))
    try:
        handle = mpi_lib.horovod_torch_join(output, device)
    except RuntimeError as e:
        raise HorovodInternalError(e)

    _handle_map[handle] = (None, output)

    return synchronize(handle).item()

def barrier(process_set=global_process_set):
    """
    A function that acts as a simple sychronization point for ranks specified
    in the given process group(default to global group). Ranks that reach
    this function call will stall until all other ranks have reached.

    Arguments:
        process_set: Process set object to limit this operation to a subset of
                     Horovod processes. Default is the global process set.
    """

    try:
        handle = mpi_lib.horovod_torch_barrier(process_set.process_set_id)
    except RuntimeError as e:
        raise HorovodInternalError(e)

    _handle_map[handle] = (None, None)

    synchronize(handle)
